import os
import json
import numpy as np
from keras_cv_attention_models import backend
from keras_cv_attention_models.backend import callbacks


class CosineLrScheduler(callbacks.Callback):
    """
    >>> from keras_cv_attention_models.imagenet import callbacks
    >>> aa = callbacks.CosineLrScheduler(0.1, first_restart_step=5, steps_per_epoch=10, m_mul=0.5)
    >>> bb = [aa.on_train_batch_begin(ii) for ii in range(350)]
    >>> plt.plot(bb)
    >>> plt.grid(True)
    """

    def __init__(self, lr_base, first_restart_step, steps_per_epoch=-1, m_mul=0.5, t_mul=2.0, lr_min=1e-5, lr_warmup=-1, warmup_steps=0, cooldown_steps=0):
        super(CosineLrScheduler, self).__init__()
        self.lr_base, self.m_mul, self.t_mul, self.lr_min, self.steps_per_epoch = lr_base, m_mul, t_mul, lr_min, steps_per_epoch
        self.first_restart_step, self.warmup_steps, self.cooldown_steps, self.lr_warmup = first_restart_step, warmup_steps, cooldown_steps, lr_warmup
        self.init_step_num, self.cur_epoch, self.is_cooldown_epoch, self.previous_cooldown_steps, self.alpha = 0, 0, False, 0, lr_min / lr_base

        self.is_built = False
        if steps_per_epoch != -1:
            self.build(steps_per_epoch)

    def cosine_decay(self, step):
        factor = 0.5 * (1 + np.cos(np.pi * (np.minimum(step, self.first_restart_batch_step) / self.first_restart_batch_step)))
        return ((1 - self.alpha) * factor + self.alpha) * self.lr_base

    def cosine_decay_restarts(self, step):
        cur_stage = (self.restart_steps > step).argmax()
        cur_lr_base = self.lr_base * (self.m_mul**cur_stage)
        cur_total_decay_steps = self.first_restart_batch_step * (self.t_mul**cur_stage)
        cur_alpha = self.lr_min / cur_lr_base
        step -= 0 if cur_stage == 0 else self.restart_steps[cur_stage - 1]
        factor = 0.5 * (1 + np.cos(np.pi * (step / cur_total_decay_steps)))
        return ((1 - cur_alpha) * factor + cur_alpha) * cur_lr_base

    def build(self, steps_per_epoch=-1):
        # print(">>>> steps_per_epoch:", steps_per_epoch)
        if steps_per_epoch != -1:
            self.steps_per_epoch = steps_per_epoch

        self.first_restart_batch_step = self.first_restart_step * self.steps_per_epoch
        alpha = self.lr_min / self.lr_base
        if self.lr_min == self.lr_base * self.m_mul:  # Without restart
            self.schedule = self.cosine_decay
            self.cooldown_steps_start, self.cooldown_steps_end = np.array([]), np.array([])
        else:
            self.schedule = self.cosine_decay_restarts
            batch_steps_each_stage = [self.first_restart_batch_step * (self.t_mul**stage) for stage in range(5)]
            self.restart_steps = np.array([int(sum(batch_steps_each_stage[:stage])) for stage in range(1, 5)])
            self.cooldown_steps_start = np.array([int(self.restart_steps[stage] / self.steps_per_epoch + self.cooldown_steps * stage) for stage in range(4)])
            self.cooldown_steps_end = np.array([ii + self.cooldown_steps for ii in self.cooldown_steps_start])

        if self.warmup_steps != 0:
            self.warmup_batch_steps = self.warmup_steps * self.steps_per_epoch
            self.lr_warmup = self.lr_warmup if self.lr_warmup > 0 else self.lr_min
            self.warmup_lr_func = lambda ii: self.lr_warmup + (self.lr_base - self.lr_warmup) * ii / self.warmup_batch_steps
        else:
            self.warmup_batch_steps = 0
        self.is_built = True

    def get_lr_with_warmup_cooldown(self, global_iterNum):
        if global_iterNum < self.warmup_batch_steps:
            lr = self.warmup_lr_func(global_iterNum)
        elif self.is_cooldown_epoch:
            lr = self.lr_min  # cooldown
        else:
            # lr = self.schedule(global_iterNum - self.warmup_batch_steps - self.previous_cooldown_steps)
            lr = self.schedule(global_iterNum - self.previous_cooldown_steps)
        return lr

    def on_epoch_begin(self, cur_epoch, logs=None):
        if not self.is_built:
            self.build()

        self.init_step_num = int(self.steps_per_epoch * cur_epoch)
        self.cur_epoch = cur_epoch

        if self.cooldown_steps_end.shape[0] != 0:
            cooldown_end_pos = (self.cooldown_steps_end > cur_epoch).argmax()
            self.previous_cooldown_steps = self.cooldown_steps * cooldown_end_pos * self.steps_per_epoch
            if cur_epoch >= self.cooldown_steps_end[cooldown_end_pos] - self.cooldown_steps:
                self.is_cooldown_epoch = True
            else:
                self.is_cooldown_epoch = False
        lr = self.get_lr_with_warmup_cooldown(self.init_step_num)
        print("\nLearning rate for iter {} is {}, global_iterNum is {}".format(self.cur_epoch + 1, lr, self.init_step_num))

    def on_train_batch_begin(self, iterNum, logs=None):
        lr = self.get_lr_with_warmup_cooldown(iterNum + self.init_step_num)
        if self.model is not None:
            # self.set_value(self.model.optimizer.lr, lr)
            if hasattr(self.model.optimizer, "param_groups"):
                for param_group in self.model.optimizer.param_groups:
                    param_group["lr"] = lr
            else:
                self.model.optimizer.learning_rate = lr
        return lr


class CosineLrSchedulerEpoch(callbacks.Callback):
    """
    >>> from keras_cv_attention_models.imagenet import callbacks
    >>> aa = callbacks.CosineLrSchedulerEpoch(0.1, first_restart_step=50, m_mul=0.5)
    >>> bb = [aa.on_epoch_begin(ii) for ii in range(350)]
    >>> plt.plot(bb)
    >>> plt.grid(True)
    """

    def __init__(self, lr_base, first_restart_step, m_mul=0.5, t_mul=2.0, lr_min=1e-6, lr_warmup=-1, warmup_steps=0, cooldown_steps=0):
        super(CosineLrSchedulerEpoch, self).__init__()
        self.warmup_steps, self.cooldown_steps, self.lr_min, self.lr_base, self.alpha = warmup_steps, cooldown_steps, lr_min, lr_base, lr_min / lr_base
        self.first_restart_step, self.m_mul, self.t_mul = first_restart_step, m_mul, t_mul

        if lr_min == lr_base * m_mul:
            self.schedule = self.cosine_decay
            self.cooldown_steps_start, self.cooldown_steps_end = np.array([]), np.array([])
        else:
            self.schedule = self.cosine_decay_restarts
            epoch_steps_each_stage = [first_restart_step * (t_mul**stage) for stage in range(5)]
            self.restart_steps = np.array([int(sum(epoch_steps_each_stage[:stage])) for stage in range(1, 5)])
            self.cooldown_steps_start = np.array([int(self.restart_steps[stage] + cooldown_steps * stage) for stage in range(4)])
            self.cooldown_steps_end = np.array([ii + cooldown_steps for ii in self.cooldown_steps_start])

        if warmup_steps != 0:
            self.lr_warmup = lr_warmup if lr_warmup > 0 else lr_min
            self.warmup_lr_func = lambda ii: self.lr_warmup + (lr_base - self.lr_warmup) * ii / warmup_steps

    def cosine_decay(self, step):
        factor = 0.5 * (1 + np.cos(np.pi * (np.minimum(step, self.first_restart_step) / self.first_restart_step)))
        return ((1 - self.alpha) * factor + self.alpha) * self.lr_base

    def cosine_decay_restarts(self, step):
        cur_stage = (self.restart_steps > step).argmax()
        cur_lr_base = self.lr_base * (self.m_mul**cur_stage)
        cur_total_decay_steps = self.first_restart_step * (self.t_mul**cur_stage)
        cur_alpha = self.lr_min / cur_lr_base
        step -= 0 if cur_stage == 0 else self.restart_steps[cur_stage - 1]
        factor = 0.5 * (1 + np.cos(np.pi * (step / cur_total_decay_steps)))
        return ((1 - cur_alpha) * factor + cur_alpha) * cur_lr_base

    def on_epoch_begin(self, epoch, logs=None):
        if epoch < self.warmup_steps:
            lr = self.warmup_lr_func(epoch)
        elif self.cooldown_steps_end.shape[0] != 0:
            cooldown_end_pos = (self.cooldown_steps_end > epoch).argmax()
            if epoch >= self.cooldown_steps_end[cooldown_end_pos] - self.cooldown_steps:
                lr = self.lr_min  # cooldown
            else:
                # lr = self.schedule(epoch - self.warmup_steps)
                lr = self.schedule(epoch - self.cooldown_steps * cooldown_end_pos)
        else:
            lr = self.schedule(epoch)

        if self.model is not None:
            # self.set_value(self.model.optimizer.lr, lr)
            if hasattr(self.model.optimizer, "param_groups"):
                for param_group in self.model.optimizer.param_groups:
                    param_group["lr"] = lr
            else:
                self.model.optimizer.learning_rate = lr
        print("\nLearning rate for iter {} is {}".format(epoch + 1, lr))
        return lr


class LearningRateScheduler(callbacks.Callback):
    def __init__(self, schedule, verbose=0):
        super().__init__()
        self.schedule, self.verbose = schedule, verbose

    def on_epoch_begin(self, epoch, logs=None):
        lr = self.schedule(epoch)
        if self.model is not None:
            if hasattr(self.model.optimizer, "param_groups"):
                for param_group in self.model.optimizer.param_groups:
                    param_group["lr"] = lr
            else:
                self.model.optimizer.learning_rate = lr
        print("\nLearning rate for iter {} is {}".format(epoch + 1, lr))
        return lr

    def on_epoch_end(self, epoch, logs=None):
        logs = logs or {}
        lr = self.model.optimizer.lr
        logs["lr"] = lr.value() if hasattr(lr, "value") else lr


def constant_scheduler(epoch, lr_base, lr_decay_steps, decay_rate=0.1, warmup_steps=0):
    if epoch < warmup_steps:
        lr = lr_base * (epoch + 1) / (warmup_steps + 1)
    else:
        # epoch -= warmup_steps
        lr = lr_base * decay_rate ** np.sum(epoch >= np.array(lr_decay_steps))
    print("\nLearning rate for iter {} is {}".format(epoch + 1, lr))
    return lr


def exp_scheduler(epoch, lr_base=0.1, decay_step=1, decay_rate=0.9, lr_min=0, warmup_steps=0):
    if epoch < warmup_steps:
        lr = (lr_base - lr_min) * (epoch + 1) / (warmup_steps + 1)
    else:
        # epoch -= warmup_steps
        lr = lr_base * decay_rate ** (epoch / decay_step)
        lr = lr if lr > lr_min else lr_min
    # print("Learning rate for iter {} is {}".format(epoch + 1, lr))
    return lr


class OptimizerWeightDecay(callbacks.Callback):
    def __init__(self, lr_base, wd_base, is_lr_on_batch=False):
        from tensorflow.keras import backend as K

        super(OptimizerWeightDecay, self).__init__()
        self.wd_m = wd_base / lr_base
        self.lr_base, self.wd_base = lr_base, wd_base
        # self.model.optimizer.weight_decay = lambda: wd_m * self.model.optimizer.lr
        self.is_lr_on_batch = is_lr_on_batch
        if is_lr_on_batch:
            self.on_train_batch_begin = self.__update_wd__
        else:
            self.on_epoch_begin = self.__update_wd__

        self.get_value, self.set_value = K.get_value, K.set_value

    def __update_wd__(self, step, log=None):
        if self.model is not None:
            wd = self.wd_m * self.get_value(self.model.optimizer.lr)
            # wd = self.wd_base * K.get_value(self.model.optimizer.lr)
            self.set_value(self.model.optimizer.weight_decay, wd)
        # wd = self.model.optimizer.weight_decay
        if not self.is_lr_on_batch or step == 0:
            print("Weight decay is {}".format(wd))


class MyHistory(callbacks.Callback):
    def __init__(self, initial_file=None):
        super(MyHistory, self).__init__()
        if initial_file and os.path.exists(initial_file):
            with open(initial_file, "r") as ff:
                self.history = json.load(ff)
        else:
            self.history = {}
        self.initial_file = initial_file

    def on_epoch_end(self, epoch, logs=None):
        logs = logs or {}
        lr = logs.pop("lr", float("nan"))
        if hasattr(self.model, "optimizer") and hasattr(self.model.optimizer, "lr"):
            lr = self.model.optimizer.lr
        elif hasattr(self.model, "optimizer") and hasattr(self.model.optimizer, "param_groups"):
            lr = self.model.optimizer.param_groups[0]["lr"]
        lr = lr.value() if hasattr(lr, "value") else lr
        lr = lr.item() if hasattr(lr, "item") else lr
        self.history.setdefault("lr", []).append(float(lr))

        for k, v in logs.items():
            k = "accuracy" if "accuracy" in k else k
            self.history.setdefault(k, []).append(float(v))

        if hasattr(self.model, "losses") and len(self.model.losses) != 0:  # Has regular_loss
            # regular_loss = functional.reduce_sum(self.model.losses).numpy()
            regular_loss = np.sum(self.model.losses)
            self.history.setdefault("regular_loss", []).append(float(regular_loss))
            self.history["loss"][-1] -= regular_loss
            if "val_loss" in self.history:
                self.history["val_loss"][-1] -= regular_loss

        model_history = self.model.history.history if hasattr(self.model.history, "history") else self.model.history
        if "val_ap_ar" in model_history:  # save coco val_ap_ar
            self.history.setdefault("val_ap_ar", []).append(model_history["val_ap_ar"][-1])

        if self.initial_file:
            with open(self.initial_file, "w") as ff:
                json.dump(self.history, ff)

    def print_hist(self):
        print("{")
        for kk, vv in self.history.items():
            print("  '%s': %s," % (kk, vv))
        print("}")


class MyCheckpoint(callbacks.Callback):
    """Save latest and best one"""

    def __init__(self, basic_save_name, monitor="val_acc", mode="auto", save_path="checkpoints"):
        super(MyCheckpoint, self).__init__()
        self.suffix = ".pt" if backend.is_torch_backend else ".h5"  # not saving .keras currently, reading weights from it is slower..
        # if backend.is_torch_backend:
        #     self.suffix = ".pt"
        # else:
        #     try:
        #         is_tf2_13 = int(tf.__version__.split(".")[0]) >= 2 and int(tf.__version__.split(".")[1]) >= 13
        #     except:
        #         is_tf2_13 = False
        #     self.suffix = ".keras" if is_tf2_13 else ".h5"
        self.basic_save_name, self.mode, self.save_path = basic_save_name, mode, save_path
        self.__init_monitor_strategy__(monitor)

    def __init_monitor_strategy__(self, monitor):
        import re

        self.monitor = monitor
        monitor_save_name = self.basic_save_name + "_epoch_{}_" + monitor + "_{}" + self.suffix
        self.monitor_save_re = re.compile(monitor_save_name.format("\d*", "[\d\.]*"))
        self.monitor_save = os.path.join(self.save_path, monitor_save_name)
        self.latest_save = os.path.join(self.save_path, self.basic_save_name + "_latest" + self.suffix)
        self.is_better = (lambda cur, pre: cur <= pre) if self.mode == "min" or "loss" in monitor else (lambda cur, pre: cur >= pre)
        self.pre_best = 1e5 if self.mode == "min" or "loss" in monitor else -1e5

    def on_epoch_end(self, epoch, logs={}):
        # tf.print(">>>> Save latest to:", self.latest_save)
        # print(">>>> logs:", logs)
        if self.model is not None:
            print("\n>>>> Save latest to:", self.latest_save)
            self.model.save(self.latest_save)
        if self.monitor is not None and self.monitor not in logs:
            all_val_acc = [ii for ii in logs.keys() if "val" in ii and "acc" in ii]
            all_val_loss = [ii for ii in logs.keys() if "val" in ii and "loss" in ii]
            if len(all_val_acc) > 0:
                self.__init_monitor_strategy__(all_val_acc[0])
            elif len(all_val_loss) > 0:
                self.__init_monitor_strategy__(all_val_loss[0])
            else:
                # self.__init_monitor_strategy__("loss")
                self.monitor = None  # Not saving if not using eval dataset

        cur_monitor_val = logs.get(self.monitor, self.pre_best)
        if self.monitor is not None and self.is_better(cur_monitor_val, self.pre_best):
            self.pre_best = cur_monitor_val
            # pre_monitor_saves = tf.io.gfile.glob(self.monitor_save_re)
            pre_monitor_saves = [ii for ii in os.listdir(self.save_path) if self.monitor_save_re.match(ii)]
            # tf.print(">>>> pre_monitor_saves:", pre_monitor_saves)
            if len(pre_monitor_saves) != 0:
                os.remove(os.path.join(self.save_path, pre_monitor_saves[0]))
            monitor_save = self.monitor_save.format(epoch + 1, "{:.4f}".format(cur_monitor_val))
            print(">>>> Save best to:", monitor_save)
            if self.model is not None:
                self.model.save(monitor_save)




from keras_cv_attention_models.common_layers import (
    activation_by_name,
    anti_alias_downsample,
    batchnorm_with_activation,
    conv2d_no_bias,
    CompatibleExtractPatches,
    dense_no_bias,
    depthwise_conv2d_no_bias,
    drop_block,
    drop_connect_rates_split,
    EvoNormalization,
    eca_module,
    fold_by_conv2d_transpose,
    global_context_module,
    group_norm,
    hard_swish,
    layer_norm,
    make_divisible,
    output_block,
    se_module,
    addaptive_pooling_2d,
    PreprocessInput,
    add_pre_post_process,
)
from keras_cv_attention_models.aotnet.aotnet import aot_stack, aot_block, deep_stem, quad_stem, tiered_stem
from keras_cv_attention_models.efficientnet.efficientnet_v2 import inverted_residual_block
from keras_cv_attention_models.resnest.resnest import rsoftmax, split_attention_conv2d
from keras_cv_attention_models.mlp_family.mlp_mixer import mlp_block, mlp_mixer_block
from keras_cv_attention_models.mlp_family.res_mlp import ChannelAffine
from keras_cv_attention_models.mlp_family.gated_mlp import spatial_gating_block
from keras_cv_attention_models.mlp_family.wave_mlp import phase_aware_token_mixing
from keras_cv_attention_models.volo.volo import outlook_attention, outlook_attention_simple, BiasLayer, PositionalEmbedding, ClassToken, MixupToken
from keras_cv_attention_models.gpt2.gpt2 import PositionalIndex, CausalMask
from keras_cv_attention_models.llama2.llama2 import PositionalEncodingFourierRot1D, RMSNorm
from keras_cv_attention_models.beit.beit import (
    PositionalEncodingFourierRot,
    MultiHeadRelativePositionalEmbedding,
    HeadInitializer,
    PatchConv2DWithResampleWeights,
    qkv_to_multi_head_channels_last_format,
    scaled_dot_product_attention,
)
from keras_cv_attention_models.botnet.botnet import RelativePositionalEmbedding, mhsa_with_relative_position_embedding
from keras_cv_attention_models.cotnet.cotnet import cot_attention
from keras_cv_attention_models.coat.coat import ConvPositionalEncoding, ConvRelativePositionalEncoding
from keras_cv_attention_models.halonet.halonet import halo_attention
from keras_cv_attention_models.convnext.convnext import global_response_normalize, add_with_layer_scale_and_drop_block
from keras_cv_attention_models.levit.levit import MultiHeadPositionalEmbedding, mhsa_with_multi_head_position, mhsa_with_multi_head_position_and_strides
from keras_cv_attention_models.nat.nat import MultiHeadRelativePositionalKernelBias, neighborhood_attention
from keras_cv_attention_models.nfnets.nfnets import ScaledStandardizedConv2D, ZeroInitGain
from keras_cv_attention_models.coatnet.coatnet import mhsa_with_multi_head_relative_position_embedding
from keras_cv_attention_models.cmt.cmt import light_mhsa_with_multi_head_relative_position_embedding, BiasPositionalEmbedding
from keras_cv_attention_models.uniformer.uniformer import multi_head_self_attention
from keras_cv_attention_models.davit.davit import (
    multi_head_self_attention_channel,
    window_attention,
    window_partition,
    window_reverse,
    pad_to_divisible_by_window_size,
    reverse_padded_for_window_size,
)
from keras_cv_attention_models.edgenext.edgenext import PositionalEncodingFourier, cross_covariance_attention
from keras_cv_attention_models.efficientvit.efficientvit_m import cascaded_mhsa_with_multi_head_position
from keras_cv_attention_models.mobilevit.mobilevit import linear_self_attention
from keras_cv_attention_models.swin_transformer_v2.swin_transformer_v2 import (
    ExpLogitScale,
    MlpPairwisePositionalEmbedding,
    shifted_window_attention,
    WindowAttentionMask,
    window_mhsa_with_pair_wise_positional_embedding,
)
from keras_cv_attention_models.hornet.hornet import ComplexDense, global_local_filter, gnconv
from keras_cv_attention_models.gpvit.gpvit import PureWeigths
from keras_cv_attention_models.pvt.pvt import mlp_block_with_depthwise_conv

CompatibleExtractPatches.__doc__ = """
For issue https://github.com/leondgarse/keras_cv_attention_models/issues/8,
Perform `tf.image.extract_patches` using `Conv2D` for TPU devices. Also for TFLite conversion.

input: `[batch, height, width, channel]`.
output (compressed=True): `[batch, height // strides,  width // strides, height_kernel * width_kernel * channel]`.
output (compressed=False): `[batch, height // strides,  width // strides, height_kernel, width_kernel, channel]`.

Args:
  sizes: could be `tf.image.extract_patches` sizes format `[1, 3, 3, 1]`, or `Conv2D` kernel_size format `3`.
  strides: could be `tf.image.extract_patches` strides format `[1, 2, 2, 1]`, or `Conv2D` strides format `2`.
  rates: could be `tf.image.extract_patches` rates format `[1, 1, 1, 1]`, or `Conv2D` dilation_rate format `1`.
  padding: "valid" or "same", will perform padding in PyTorch way if "same".
  compressed: boolean value if compress extracted `height_kernel`, `width_kernel`, `channel` into 1 dimension.
  force_conv: force using `Conv2D` instead of `tf.image.extract_patches`.

Examples:

>>> from keras_cv_attention_models import attention_layers
>>> aa = tf.ones([1, 64, 27, 192])
>>> print(attention_layers.CompatibleExtractPatches(sizes=3, strides=2)(aa).shape)
# (1, 32, 14, 1728)
>>> print(attention_layers.CompatibleExtractPatches(sizes=3, strides=2, compressed=False)(aa).shape)
# (1, 32, 14, 3, 3, 192)

# `Conv2D` version Performs slower than `extract_patches`.
>>> cc = attention_layers.CompatibleExtractPatches(sizes=3, strides=2, force_conv=True)
>>> cc(aa).shape  # init run
>>> %timeit cc(aa)
# 772 µs ± 6.71 µs per loop (mean ± std. dev. of 7 runs, 1000 loops each)
>>> %timeit tf.image.extract_patches(aa, [1, 3, 3, 1], [1, 2, 2, 1], [1, 1, 1, 1], padding='same')
# 108 µs ± 153 ns per loop (mean ± std. dev. of 7 runs, 10000 loops each)
"""

fold_by_conv2d_transpose.__doc__ = """
Fold function like `torch.nn.Fold` using `Conv2DTranspose`.

input (compressed=True): `[batch, height // strides,  width // strides, height_kernel * width_kernel * channel]`.
input (compressed=False): `[batch, height // strides,  width // strides, height_kernel, width_kernel, channel]`.
output: `[batch, height, width, channel]`.

Args:
  patches: input tensor.
  output_shape: specific output shape in format `(height, width)`.
      Default `None` will just cut `padded` by `out[:, paded:-paded, paded:-paded, :]`.
  kernel_size: same as `Conv2DTranspose` kernel_size. Default `3`.
  strides: same as `Conv2DTranspose` strides. Default `2`.
  dilation_rate: same as `Conv2DTranspose` dilation_rate. Default `1`.
  padding: "valid" or "same", indicates if `patches` is generated from a padded input. Default "same".
  compressed: boolean value if `patches` last dimension is a compressed of `height_kernel`, `width_kernel`, `channel`.
      Default "auto" means auto judge from `patches` shape length.

Examples:

>>> inputs = np.random.uniform(size=[1, 64, 27, 192]).astype("float32")
>>> kernel_size, strides = 3, 2

>>> # ==== Torch unfold - fold ====
>>> import torch
>>> torch_inputs = torch.from_numpy(inputs).permute(0, 3, 1, 2)  # NCHW
>>> unfold = torch.nn.Unfold(kernel_size=kernel_size, dilation=1, padding=1, stride=strides)
>>> fold = torch.nn.Fold(output_size=torch_inputs.shape[2:4], kernel_size=kernel_size, dilation=1, padding=1, stride=strides)
>>> torch_patches = unfold(torch_inputs)
>>> torch_fold = fold(torch_patches)

>>> # ==== TF unfold - fold ====
>>> from keras_cv_attention_models import attention_layers
>>> tf_patches = attention_layers.CompatibleExtractPatches(kernel_size, strides)(inputs)
>>> tf_fold = attention_layers.fold_by_conv2d_transpose(tf_patches, output_shape=inputs.shape[1:], kernel_size=kernel_size, strides=strides)
>>> print(f"{np.allclose(tf_fold, torch_fold.permute([0, 2, 3, 1])) = }")
# np.allclose(tf_fold, torch_fold.permute([0, 2, 3, 1])) = True

>>> # ==== TF extract_patches ====
>>> pad = kernel_size // 2
>>> pad_inputs = tf.pad(inputs, [[0, 0], [pad, pad], [pad, pad], [0, 0]])
>>> patches = tf.image.extract_patches(pad_inputs, [1, kernel_size, kernel_size, 1], [1, strides, strides, 1], [1, 1, 1, 1], padding='valid')
>>> print(f"{np.allclose(tf_patches, patches) = }")
# np.allclose(tf_patches, patches) = True
"""

EvoNormalization.__doc__ = """
Keras implementation of [evonorm](https://github.com/tensorflow/tpu/blob/master/models/official/resnet/resnet_model.py).
Paper [PDF 2004.02967 Evolving Normalization-Activation Layers](https://arxiv.org/pdf/2004.02967.pdf).

Args:
  nonlinearity: boolean value if apply non-linearity calculation, default True.
  num_groups: groups for paper `EVONORM_S0` mode, value `<=0` for `EVONORM_B0` mode. Dafult -1.
      - EVONORM_B0: nonlinearity=True, num_groups=-1
      - EVONORM_S0: nonlinearity=True, num_groups > 0
      - EVONORM_B0 / EVONORM_S0 linearity: nonlinearity=False, num_groups=-1
      - EVONORM_S0A linearity: nonlinearity=False, num_groups > 0.
        Defined in https://github.com/rwightman/pytorch-image-models/blob/main/timm/models/layers/evo_norm.py#L239.
  zero_gamma: boolean value if init gamma as 0, default False for init as 1.
  momentum: momentum for `EVONORM_B0` mode, default 0.99.
  epsilon: small constant avoid zero dividing, default 0.001.
  data_format: one of ["channels_first", "channels_last"]. Default "auto" for using `K.image_data_format()`.

Examples:
>>> from keras_cv_attention_models import attention_layers
>>> # EVONORM_B0
>>> aa = attention_layers.EvoNormalization()
>>> print(aa(tf.ones([2, 32, 32, 48])).shape)
>>> # (2, 32, 32, 48)
>>> print({ii.name.split("/")[-1]: ii.shape for ii in aa.weights})
>>> # {'gamma:0': TensorShape([1, 1, 1, 48]), 'beta:0': TensorShape([1, 1, 1, 48]), 'vv:0': TensorShape([1, 1, 1, 48]), 'moving_variance:0': TensorShape([1, 1, 1, 48])}

>>> # EVONORM_B0 linearity
>>> aa = attention_layers.EvoNormalization(nonlinearity=False)
>>> print(aa(tf.ones([2, 32, 32, 56, 48])).shape)
>>> # (2, 32, 32, 56, 48)
>>> print({ii.name.split("/")[-1]: ii.shape for ii in aa.weights})
>>> # {'gamma:0': TensorShape([1, 1, 1, 1, 48]), 'beta:0': TensorShape([1, 1, 1, 1, 48]), 'moving_variance:0': TensorShape([1, 1, 1, 1, 48])}

>>> # EVONORM_S0, actual using __num_groups__ is made divisible by channel dimension.
>>> aa = attention_layers.EvoNormalization(nonlinearity=True, num_groups=32)
>>> print(aa(tf.ones([2, 32, 56])).shape, aa.__num_groups__) # print actual using __num_groups__
>>> # (2, 32, 56) 28
>>> print({ii.name.split("/")[-1]: ii.shape for ii in aa.weights})
>>> # {'gamma:0': TensorShape([1, 1, 56]), 'beta:0': TensorShape([1, 1, 56]), 'vv:0': TensorShape([1, 1, 56])}

>>> # EVONORM_S0, force "channels_first".
>>> aa = attention_layers.EvoNormalization(nonlinearity=True, num_groups=32, data_format="channels_first")
>>> print(aa(tf.ones([2, 32, 56])).shape, aa.__num_groups__) # print actual using __num_groups__
>>> # (2, 32, 56) 32
>>> print({ii.name.split("/")[-1]: ii.shape for ii in aa.weights})
>>> # {'gamma:0': TensorShape([1, 32, 1]), 'beta:0': TensorShape([1, 32, 1]), 'vv:0': TensorShape([1, 32, 1])}

>>> # EVONORM_S0A linearity
>>> aa = attention_layers.EvoNormalization(nonlinearity=False, num_groups=16)
>>> print(aa(tf.ones([2, 48])).shape)
>>> # (2, 48)
>>> print({ii.name.split("/")[-1]: ii.shape for ii in aa.weights})
>>> # {'gamma:0': TensorShape([1, 48]), 'beta:0': TensorShape([1, 48])}
"""



from keras_cv_attention_models.yolox.yolox import (
    CSPDarknet,
    YOLOX,
    YOLOXNano,
    YOLOXTiny,
    YOLOXS,
    YOLOXM,
    YOLOXL,
    YOLOXX,
)

__head_doc__ = """
Keras implementation of [Megvii-BaseDetection/YOLOX](https://github.com/Megvii-BaseDetection/YOLOX).
Paper [Paper 2107.08430 YOLOX: Exceeding YOLO Series in 2021](https://arxiv.org/pdf/2107.08430.pdf).

Args:
  backbone: backbone model, could be any model with pyramid stage structure.
      Default None for CSPDarknet with depth_mul={depth_mul}, width_mul={width_mul}.
"""

__tail_doc__ = """  features_pick: specific `layer names` or `pyramid feature indexes` from backbone model.
        Default `[-3, -2, -1]` means using the last 3 pyramid feature output from backbone.
  regression_len: bbox output len, typical value is 4, for yolov8 reg_max=16 -> regression_len = 16 * 4 == 64.
  anchors_mode: one of ["efficientdet", "anchor_free", "yolor"], controls which anchor to use.
      - efficientdet anchors default settings: use_object_scores=False, num_anchors=9, anchor_scale=4,
          aspect_ratios=[1, 2, 0.5], num_scales=3, grid_zero_start=False.
      - anchor_free default settings: use_object_scores=True, num_anchors=1, anchor_scale=1,
          aspect_ratios=[1], num_scales=1, grid_zero_start=True.
      - yolor default settings: use_object_scores=True, num_anchors=3.
      - yolov8 default settings: use_object_scores=False, num_anchors=1, anchor_scale=1,
          aspect_ratios=[1], num_scales=1, grid_zero_start=False.
      Default "anchor_free".
  num_anchors: number of anchors for a single grid point, should be same with dataset used value.
      Default "auto" means: anchors_mode=="anchor_free" / "yolov8" -> 1, anchors_mode=="yolor" -> 3, else 9.
  use_object_scores: bollean value if model header output includes `object_scores`.
      Default "auto" means: True if anchors_mode=="anchor_free" or anchors_mode=="yolor", else False.
  num_classes: total output classes. Set `0` to disable `classifier` output. Default 80 for COCO.
  input_shape: input shape if backbone is None, else will use input_shape from backbone.
  activation: activation used in whole model, default `swish`. Default "swish".
  freeze_backbone: set `True` for `backbone.trainable = False`. Default `False`.
  pretrained: one of `None` (random initialization) or 'coco' (pre-training on COCO).
      Will try to download and load pre-trained model weights if not None. Default `coco`.
  pyramid_levels_min: anchors inititial parameter for model prediction, not affecting model architecture. Default `3`.
      pyramid_levels_max is calculated as `pyramid_levels_min + len(features_pick) - 1`.
  anchor_scale: anchors inititial parameter for model prediction, not affecting model architecture.
      Default "auto" means: 1 if (anchors_mode=="anchor_free" or anchors_mode=="yolor"), else 4.
  rescale_mode: model precessing input, not for model structure. Defulat "raw" means input value in range `[0, 255]`.

Returns:
    A `keras.Model` instance.
"""

YOLOX.__doc__ = __head_doc__.format(depth_mul=1, width_mul=1) + """  depth_mul: CSPDarknet backbone and FPN blocks depth expansion ratio.
      - For CSPDarknet, base_depth = max(round(depth_mul * 3), 1)
      - For FPN blocks, csp_depth = max(round(depth_mul * 3), 1)
  width_mul: CSPDarknet backbone and output header width expansion ratio.
      - For CSPDarknet, base_channels = int(width_mul * 64)
      - For output header, out_channel = int(256 * width_mul)
      Default -1 means: `min([ii.shape[-1] for ii in features]) / 256` for custom backbones.
  use_depthwise_conv: boolean value if using additional depthwise conv. True for YOLOXNano, False for others.
  model_name: string, model name.
""" + __tail_doc__ + """
Model architectures:
  | Model     | Params | FLOPs   | Input | COCO val AP | test AP |
  | --------- | ------ | ------- | ----- | ----------- | ------- |
  | YOLOXNano | 0.91M  | 0.53G   | 416   | 25.8        |         |
  | YOLOXTiny | 5.06M  | 3.22G   | 416   | 32.8        |         |
  | YOLOXS    | 9.0M   | 13.39G  | 640   | 40.5        | 40.5    |
  | YOLOXM    | 25.3M  | 36.84G  | 640   | 46.9        | 47.2    |
  | YOLOXL    | 54.2M  | 77.76G  | 640   | 49.7        | 50.1    |
  | YOLOXX    | 99.1M  | 140.87G | 640   | 51.5        | 51.5    |
"""

YOLOXNano.__doc__ = __head_doc__.format(depth_mul=0.33, width_mul=0.25) + __tail_doc__
YOLOXTiny.__doc__ = __head_doc__.format(depth_mul=0.33, width_mul=0.375) + __tail_doc__
YOLOXS.__doc__ = __head_doc__.format(depth_mul=0.33, width_mul=0.5) + __tail_doc__
YOLOXM.__doc__ = __head_doc__.format(depth_mul=0.67, width_mul=0.75) + __tail_doc__
YOLOXL.__doc__ = __head_doc__.format(depth_mul=1.0, width_mul=1.0) + __tail_doc__
YOLOXX.__doc__ = __head_doc__.format(depth_mul=1.33, width_mul=1.25) + __tail_doc__



from tensorflow.nn import *
from tensorflow.math import *

from tensorflow import (
    abs,
    cast,
    clip_by_value,
    complex,
    concat,
    convert_to_tensor,
    expand_dims,
    gather,
    gather_nd,
    linspace,
    map_fn,
    matmul,
    norm,
    pad,
    print,
    range,
    repeat,
    reshape,
    shape,
    sign,
    split,
    squeeze,
    stack,
    tensor_scatter_nd_update,
    tile,
    transpose,
    unstack,
    zeros,
)
from tensorflow.image import resize, extract_patches, non_max_suppression_with_scores
from tensorflow.signal import irfft2d, rfft2d



""" CLIP tokenizer

Copied from https://github.com/mlfoundations/open_clip/blob/main/src/open_clip/tokenizer.py
Copied from https://github.com/openai/CLIP. Originally MIT License, Copyright (c) 2021 OpenAI.
"""
import os
import html
from functools import lru_cache
import numpy as np


os.environ["TOKENIZERS_PARALLELISM"] = "false"  # https://stackoverflow.com/q/62691279

DEFAULT_SOT = "<|startoftext|>"
DEFAULT_EOT = "<|endoftext|>"

# optional keys: {"limit_vocab_size": None, "sot": DEFAULT_SOT, "eot": DEFAULT_EOT, "is_space_first": False}
BUILDIN_TOKENIZERS = {
    # [???] limit_vocab_size = 49152-256-2 == 48894 for original bpe vocab, why this number?
    "clip": {"path": "bpe_simple_vocab_16e6.txt", "file_hash": "f83f3e2479df59e7cd1597baa03f34c8", "limit_vocab_size": 48894},
    # "<|start_of_text|>" not exists in original gpt vocab, using sot == eot here
    "gpt2": {"path": "gpt2_tokenizer.txt", "file_hash": "150fdad3c88ee8f9607ac1808ad2d321", "sot": DEFAULT_EOT, "eot": DEFAULT_EOT, "is_space_first": True},
    "llama": {"path": "llama_tokenizer.model", "file_hash": "eeec4125e9c7560836b4873b6f8e3025", "sot": "<s>", "eot": "</s>", "is_space_first": True},
    "llama2": {"path": "llama_tokenizer.model", "file_hash": "eeec4125e9c7560836b4873b6f8e3025", "sot": "<s>", "eot": "</s>", "is_space_first": True},
}


def download_tokenizer_file(name):
    from keras_cv_attention_models.backend import get_file

    config = BUILDIN_TOKENIZERS[name]
    path, file_hash = config["path"], config["file_hash"]

    url = "https://github.com/leondgarse/keras_cv_attention_models/releases/download/assets/{}".format(path)
    tokenizer_file = os.path.join(os.path.expanduser("~"), ".keras", "datasets", path)
    print(">>>> Load tokenizer from file:", tokenizer_file)
    return get_file(origin=url, file_hash=file_hash)


class SimpleTokenizer(object):
    def __init__(self, name_or_path="clip", special_tokens=None, limit_vocab_size="auto", context_length=77):
        import ftfy  # fixes text for you, importing here avoiding additional requirements if not needed
        import regex

        self.ftfy, self.regex = ftfy, regex

        self.byte_encoder = self.bytes_to_unicode()
        self.byte_decoder = {vv: kk for kk, vv in self.byte_encoder.items()}
        byte_vocab = self._init_byte_vacab_()  # Different from gpt2 and clip

        tokens, self.sot, self.eot, self.is_space_first = self._init_tokenizer_from_file_(name_or_path, limit_vocab_size)
        tokens_split = [ii.split() for ii in tokens if len(ii) != 0]  # exclude empty one from gpt2
        self.bpe_ranks = {tuple(ii): id for id, ii in enumerate(tokens_split)}
        token_vocab = ["".join(ii) for ii in tokens_split]
        vocab = [ii for ii in byte_vocab if ii not in token_vocab] + token_vocab  # filter basic byte_vocab from provided token_vocab

        special_tokens = ([self.sot] if self.sot == self.eot else [self.sot, self.eot]) + (special_tokens if special_tokens else [])
        self.cache = {t: t for t in special_tokens}
        special_regex = "|".join([ii.replace("|", "\|") for ii in special_tokens])
        self.pat = regex.compile(special_regex + r"""|'s|'t|'re|'ve|'m|'ll|'d|[\p{L}]+|[\p{N}]|[^\s\p{L}\p{N}]+""", regex.IGNORECASE)
        vocab.extend(special_tokens)

        self.decoder = dict(enumerate(vocab))
        self.encoder = {v: k for k, v in self.decoder.items()}
        self.vocab_size = len(vocab)

        self.all_special_ids = [self.encoder[t] for t in special_tokens]
        self.sot_token, self.eot_token = self.encoder[self.sot], self.encoder[self.eot]
        self.context_length = context_length

    def _init_byte_vacab_(self):
        return list(self.byte_encoder.values()) + [vv + "</w>" for vv in self.byte_encoder.values()]

    def _init_tokenizer_from_file_(self, name_or_path, limit_vocab_size):
        name_or_path = name_or_path.lower()
        if name_or_path in BUILDIN_TOKENIZERS:
            tokenizer_file = download_tokenizer_file(name_or_path)

            config = BUILDIN_TOKENIZERS[name_or_path]
            sot, eot, is_space_first = config.get("sot", DEFAULT_SOT), config.get("eot", DEFAULT_EOT), config.get("is_space_first", False)
            limit_vocab_size = config.get("limit_vocab_size", None) if limit_vocab_size == "auto" else limit_vocab_size
        else:
            tokenizer_file = name_or_path
            limit_vocab_size = None if limit_vocab_size == "auto" else limit_vocab_size
            sot, eot, is_space_first = DEFAULT_SOT, DEFAULT_EOT, False

        if tokenizer_file.endswith(".gz") or tokenizer_file.endswith(".zip"):
            import gzip

            with gzip.open(tokenizer_file) as ff:
                tokens = ff.read().decode("utf-8").split("\n")
        else:
            with open(tokenizer_file) as ff:
                tokens = ff.read().split("\n")[1:]  # exclude first line
        if limit_vocab_size:
            tokens = tokens[:limit_vocab_size]
        return tokens, sot, eot, is_space_first

    @staticmethod
    @lru_cache()
    def bytes_to_unicode():
        """
        Returns list of utf-8 byte and a corresponding list of unicode strings.
        The reversible bpe codes work on unicode strings.
        This means you need a large # of unicode characters in your vocab if you want to avoid UNKs.
        When you're at something like a 10B token dataset you end up needing around 5K for decent coverage.
        This is a significant percentage of your normal, say, 32K bpe vocab.
        To avoid that, we want lookup tables between utf-8 bytes and unicode strings.
        And avoids mapping to whitespace/control characters the bpe code barfs on.
        """
        # Without [0, 32], [127, 160]
        bs = list(range(ord("!"), ord("~") + 1)) + list(range(ord("¡"), ord("¬") + 1)) + list(range(ord("®"), ord("ÿ") + 1))
        additional = [b for b in range(2**8) if b not in bs]
        cs = bs + [ii for ii in range(2**8, 2**8 + len(additional))]
        return dict(zip(bs + additional, [chr(ii) for ii in cs]))

    @staticmethod
    def _get_pairs_(word):
        """Return set of symbol pairs in a word. get_pairs('hello') -> {('e', 'l'), ('h', 'e'), ('l', 'l'), ('l', 'o')}"""
        return set([(pre, cur) for pre, cur in zip(word[:-1], word[1:])])

    def bpe(self, token, is_first_token=False):
        token_with_space = (token if is_first_token else ("</w>" + token)) if self.is_space_first else (token + "</w>")
        if token_with_space in self.encoder:
            # self.cache[token_with_space] = token_with_space
            return token_with_space
        if token_with_space in self.cache:
            return self.cache[token_with_space]

        if self.is_space_first:
            word = tuple(token) if is_first_token else (("</w>" + token[0],) + tuple(token[1:]))  # gpt2 one
        else:
            word = tuple(token[:-1]) + (token[-1] + "</w>",)  # clip one

        pairs = self._get_pairs_(word)
        if not pairs:
            return token_with_space

        # print(f"{pairs = }")
        cur = word
        while len(cur) > 1:
            bigram = min(pairs, key=lambda pair: self.bpe_ranks.get(pair, float("inf")))
            if bigram not in self.bpe_ranks:
                break
            first, second = bigram
            new_word = []
            while len(cur) > 0:
                if first in cur:
                    next_index = cur.index(first)
                    new_word.extend(cur[:next_index])
                    cur = cur[next_index:]
                else:
                    new_word.extend(cur)
                    break

                if cur[0] == first and len(cur) > 1 and cur[1] == second:
                    new_word.append(first + second)
                    cur = cur[2:]
                else:
                    new_word.append(cur[0])
                    cur = cur[1:]
            cur = tuple(new_word)
            pairs = self._get_pairs_(cur)

        word = " ".join(cur)
        # print(f"{word = }")
        self.cache[token] = word
        return word

    def text_clean(self, text):
        text = self.ftfy.fix_text(text)
        text = html.unescape(html.unescape(text))
        return self.regex.sub(r"\s+", " ", text.strip()).strip()

    def encode(self, text, add_sot=False, add_eot=False):
        bpe_tokens = []
        text = self.text_clean(text).lower()
        is_first_token = True
        for token in self.regex.findall(self.pat, text):
            token = "".join(self.byte_encoder[b] for b in token.encode("utf-8"))
            # print(f"{token = }")
            bpe_tokens.extend(self.encoder[bpe_token] for bpe_token in self.bpe(token, is_first_token=is_first_token).split(" "))
            is_first_token = False

        if add_sot:
            bpe_tokens = [self.sot_token] + bpe_tokens
        if add_eot:
            bpe_tokens = bpe_tokens + [self.sot_token]
        return bpe_tokens

    def decode(self, tokens):
        text = "".join([self.decoder[token] for token in tokens])
        text = bytearray([self.byte_decoder[c] for c in text]).decode("utf-8", errors="replace").replace("</w>", " ")
        return text

    def __call__(self, inputs, padding_value=0):
        if isinstance(inputs, str) or isinstance(inputs, bytes):
            inputs = inputs.decode() if hasattr(inputs, "decode") else inputs
            tokens = [self.sot_token] + self.encode(inputs)[: self.context_length - 2] + [self.eot_token]
            # print(f"{tokens = }")
            return np.pad(tokens, [0, self.context_length - len(tokens)], constant_values=padding_value)
        else:
            inputs = inputs.detach() if hasattr(inputs, "detach") else inputs
            inputs = inputs.cpu() if hasattr(inputs, "cpu") else inputs
            inputs = inputs.numpy() if hasattr(inputs, "numpy") else inputs
            inputs = inputs.tolist() if hasattr(inputs, "tolist") else list(inputs)

            inputs = inputs[1 if inputs[0] == self.sot_token else 0 :]
            inputs = inputs[: inputs.index(self.eot_token) if self.eot_token in inputs else None]
            return self.decode(inputs)


class GPT2Tokenizer(SimpleTokenizer):
    def __init__(self, name_or_path="gpt2", special_tokens=None, limit_vocab_size="auto", context_length=77):
        super().__init__(name_or_path, special_tokens, limit_vocab_size, context_length)

    def _init_byte_vacab_(self):
        return list(self.byte_encoder.values())


class SentencePieceTokenizer(SimpleTokenizer):
    """
    >>> from keras_cv_attention_models.clip import tokenizer
    >>> ee = tokenizer.SentencePieceTokenizer()
    >>> print(ee(ee('hello world')))
    >>> #  hello world  <- An additional leading white space
    >>> print(ee.decode(ee.encode('hello world', add_sot=True)))
    >>> # hello world
    """

    def __init__(self, name_or_path="llama", context_length=77):
        from sentencepiece import SentencePieceProcessor

        self.name_or_path = download_tokenizer_file(name_or_path) if name_or_path in BUILDIN_TOKENIZERS else name_or_path
        self.sp_model = SentencePieceProcessor(model_file=self.name_or_path)
        assert self.sp_model.vocab_size() == self.sp_model.get_piece_size()
        self.decoder, self.scores = self._build_vocab_dict_()

        # BOS / EOS token IDs
        self.vocab_size = self.sp_model.vocab_size()
        self.sot_token, self.eot_token = self.sp_model.bos_id(), self.sp_model.eos_id()
        self.pad_token = self.sp_model.pad_id()

        self.sot = self.decoder[self.sot_token]
        self.eot = self.decoder[self.eot_token]
        self.context_length = context_length

    def _build_vocab_dict_(self):
        # https://github.com/karpathy/llama2.c/blob/master/tokenizer.py#L42-L61
        decoder, scores = {}, {}
        for id in range(self.sp_model.vocab_size()):
            token = self.sp_model.id_to_piece(id)
            score = self.sp_model.get_score(id)
            if id == self.sp_model.bos_id():
                token = "\n" + token + "\n"
            elif id == self.sp_model.eos_id():
                token = "\n" + token + "\n"
            elif len(token) == 6 and token.startswith("<0x") and token.endswith(">"):
                token = chr(int(token[3:5], 16))  # e.g. make '<0x01>' into '\x01'
            token = token.replace("▁", " ")  # sentencepiece uses this character as whitespace
            decoder[id] = token  # .encode('utf-8') # bytes of this token, utf-8 encoded
            scores[id] = score
        return decoder, scores

    def encode(self, text, add_sot=False, add_eot=False):
        tokens = self.sp_model.encode(text)
        if add_sot:
            tokens = [self.sot_token] + tokens
        if add_eot:
            tokens = tokens + [self.sot_token]
        return tokens

    # def decode_single(self, token):
    #     return self.decoder[int(tokens)]

    def decode(self, tokens):
        # Not using sp_model.decode, as leading space is omitted when decoding a single word. https://github.com/karpathy/llama2.c/pull/89
        # return self.sp_model.decode(tokens)
        tokens = tokens.tolist() if hasattr(tokens, "tolist") else tokens
        if isinstance(tokens, (list, tuple)) and len(tokens) > 1:
            return self.sp_model.decode(tokens)
            # rr = "".join([self.decoder[int(ii)] for ii in tokens])
            # return (self.sot + rr[len(self.sot) + 1 :]) if tokens[0] == self.sot_token else rr
        else:
            token = int(tokens[0]) if isinstance(tokens, (list, tuple)) else int(tokens)
            return self.decoder[token]


class TikToken(SimpleTokenizer):
    """OpenAI  tiktoken wrapper"""

    def __init__(self, encoding_name="gpt2", context_length=77):
        import tiktoken
        import ftfy  # fixes text for you, importing here avoiding additional requirements if not needed
        import regex

        self.ftfy, self.regex = ftfy, regex

        if encoding_name not in tiktoken.list_encoding_names():
            raise ValueError("[Error] encoding_name should be one of {}".format(tiktoken.list_encoding_names()))

        self.tokenizer = tiktoken.get_encoding(encoding_name)
        self.sot = self.eot = "<|endoftext|>"
        self.sot_token = self.eot_token = self.tokenizer.encode(self.eot, allowed_special={self.eot})[0]
        self.vocab_size = self.tokenizer.n_vocab
        self.name = self.encoding_name = encoding_name

        self.context_length = context_length

    def encode(self, text, add_sot=False, add_eot=False):
        text = self.text_clean(text).lower()
        tokens = self.tokenizer.encode(text)

        if add_sot:
            tokens = [self.sot_token] + tokens
        if add_eot:
            tokens = tokens + [self.sot_token]
        return tokens

    def decode(self, tokens):
        return self.tokenizer.decode(tokens)


class HuggingFaceTokenizer:
    """HuggingFace tokenizer wrapper"""

    def __init__(self, tokenizer_name: str):
        from transformers import AutoTokenizer

        self.tokenizer = AutoTokenizer.from_pretrained(tokenizer_name)

        self.context_length = 77

    def save_pretrained(self, dest):
        self.tokenizer.save_pretrained(dest)

    def __call__(self, texts, context_length: int = 77):
        # same cleaning as for default tokenizer, except lowercasing
        # adding lower (for case-sensitive tokenizers) will make it more robust but less sensitive to nuance
        if isinstance(texts, str):
            texts = [texts]
        texts = [whitespace_clean(basic_clean(text)) for text in texts]
        input_ids = self.tokenizer(
            texts,
            return_tensors="pt",
            max_length=context_length,
            padding="max_length",
            truncation=True,
        ).input_ids
        return input_ids



from keras_cv_attention_models.efficientvit.efficientvit_m import (
    EfficientViT_M,
    EfficientViT_M0,
    EfficientViT_M1,
    EfficientViT_M2,
    EfficientViT_M3,
    EfficientViT_M4,
    EfficientViT_M5,
    cascaded_mhsa_with_multi_head_position,
)
from keras_cv_attention_models.efficientvit.efficientvit_b import (
    EfficientViT_B,
    EfficientViT_B0,
    EfficientViT_B1,
    EfficientViT_B2,
    EfficientViT_B3,
    EfficientViT_L,
    EfficientViT_L0,
    EfficientViT_L1,
    EfficientViT_L2,
    EfficientViT_L3,
)

__m_head_doc__ = """
Keras implementation of [Github microsoft/Cream/EfficientViT/classification](https://github.com/microsoft/Cream/tree/main/EfficientViT/classification).
Paper [PDF 2305.07027 EfficientViT: Memory Efficient Vision Transformer with Cascaded Group Attention](https://arxiv.org/pdf/2305.07027.pdf).
"""

__tail_doc__ = """  input_shape: it should have exactly 3 inputs channels, like `(224, 224, 3)`.
  num_classes: number of classes to classify images into. Set `0` to exclude top layers.
  activation: activation used in whole model, default `relu`.
  drop_connect_rate: is used for [Deep Networks with Stochastic Depth](https://arxiv.org/abs/1603.09382).
      Can be a constant value like `0.2`,
      or a tuple value like `(0, 0.2)` indicates the drop probability linearly changes from `0 --> 0.2` for `top --> bottom` layers.
      A higher value means a higher probability will drop the deep branch.
      or `0` to disable (default).
  dropout: top dropout rate if top layers is included. Default 0.
  classifier_activation: A `str` or callable. The activation function to use on the "top" layer if `num_classes > 0`.
      Set `classifier_activation=None` to return the logits of the "top" layer.
      Default is `None`.
  use_distillation: Boolean value if output `distill_head`. Default `False`.
  pretrained: one of `None` (random initialization) or 'imagenet'.
      Will try to download and load pre-trained model weights if not None.

Returns:
    A `keras.Model` instance.
"""

EfficientViT_M.__doc__ = __m_head_doc__ + """
Args:
  num_blocks: number of blocks in each stack.
  out_channels: output channels for each stack.
  window_size: int value indicates window partition size for attention blocks.
  num_heads: int or list value indicates num_heads for attention blocks in each stack.
  key_dim: int value indicates key dim for attention blocks query and key.
  kernel_sizes: int or list value indicates kernel_size for each head in attention blocks. Length must larger than max of num_heads.
  mlp_ratio: int value indicates expand ratio for mlp blocks hidden channels.
  stem_width: output dimension for stem block. Defailt -1 for using `out_channels[0]`.
  model_name: string, model name.
""" + __tail_doc__ + """
Model architectures:
  | Model           | Params | FLOPs | Input | Top1 Acc |
  | --------------- | ------ | ----- | ----- | -------- |
  | EfficientViT_M0 | 2.35M  | 79.4M | 224   | 63.2     |
  | EfficientViT_M1 | 2.98M  | 167M  | 224   | 68.4     |
  | EfficientViT_M2 | 4.19M  | 201M  | 224   | 70.8     |
  | EfficientViT_M3 | 6.90M  | 263M  | 224   | 73.4     |
  | EfficientViT_M4 | 8.80M  | 299M  | 224   | 74.3     |
  | EfficientViT_M5 | 12.47M | 522M  | 224   | 77.1     |
"""

EfficientViT_M0.__doc__ = __m_head_doc__ + """
Args:
""" + __tail_doc__

EfficientViT_M1.__doc__ = EfficientViT_M0.__doc__
EfficientViT_M2.__doc__ = EfficientViT_M0.__doc__
EfficientViT_M3.__doc__ = EfficientViT_M0.__doc__
EfficientViT_M4.__doc__ = EfficientViT_M0.__doc__
EfficientViT_M5.__doc__ = EfficientViT_M0.__doc__

__b_head_doc__ = """
Keras implementation of [Github mit-han-lab/efficientvit](https://github.com/mit-han-lab/efficientvit).
Paper [PDF 2205.14756 EfficientViT: Lightweight Multi-Scale Attention for On-Device Semantic Segmentation](https://arxiv.org/pdf/2205.14756.pdf).
"""

EfficientViT_B.__doc__ = __b_head_doc__ + """
Args:
  num_blocks: number of blocks in each stack.
  out_channels: output channels for each stack.
  stem_width: output dimension for stem block.
  block_types: block types for each stack,
      - `conv` or any `c` / `C` starts word, means `mb_conv` block.
      - `transfrom` or any not `c` / `C` starts word, means `lite_mhsa` + `mb_conv` block.
      value could be in format like `"cctt"` or `"CCTT"` or `["conv", "conv", "transfrom", "transform"]`.
  expansions: int or list, each element in list can also be an int or list of int.
      Indicates filter expansion in each block. The larger the wider.
  is_fused: boolean or list of boolean, indicates if using fused MBConv in each stack.
  head_dimension: heads dimension for transformer block, and `num_heads = channels // head_dimension`.
  output_filters: int or list of int value indicates id using additional blocks before output.
      - int value for adding `Conv2D + BatchNorm + Activation` block, and the value is used as conv2D filters.
      - list value for adding another `Dense + LayerNorm + Activation` block, and the second value is used as Dense filters.
      - Set `0` for disable.
  model_name: string, model name.
""" + __tail_doc__ + """
Model architectures:
  | Model           | Params | FLOPs | Input | Top1 Acc |
  | --------------- | ------ | ----- | ----- | -------- |
  | EfficientViT_B0 | 3.41M  | 0.12G | 224   | 71.6 ?   |
  | EfficientViT_B1 | 9.10M  | 0.58G | 224   | 79.4     |
  | EfficientViT_B1 | 9.10M  | 0.78G | 256   | 79.9     |
  | EfficientViT_B1 | 9.10M  | 1.03G | 288   | 80.4     |
  | EfficientViT_B2 | 24.33M | 1.68G | 224   | 82.1     |
  | EfficientViT_B2 | 24.33M | 2.25G | 256   | 82.7     |
  | EfficientViT_B2 | 24.33M | 2.92G | 288   | 83.1     |
  | EfficientViT_B3 | 48.65M | 4.14G | 224   | 83.5     |
  | EfficientViT_B3 | 48.65M | 5.51G | 256   | 83.8     |
  | EfficientViT_B3 | 48.65M | 7.14G | 288   | 84.2     |
  | EfficientViT_L1 | 52.65M | 5.28G | 224   | 84.48    |
  | EfficientViT_L2 | 63.71M | 6.98G | 224   | 85.05    |
  | EfficientViT_L2 | 63.71M | 20.7G | 384   | 85.98    |
  | EfficientViT_L3 | 246.0M | 27.6G | 224   | 85.814   |
  | EfficientViT_L3 | 246.0M | 81.6G | 384   | 86.408   |
"""

EfficientViT_L.__doc__ = EfficientViT_B.__doc__

EfficientViT_B0.__doc__ = __b_head_doc__ + """
Args:
""" + __tail_doc__

EfficientViT_B1.__doc__ = EfficientViT_B0.__doc__
EfficientViT_B2.__doc__ = EfficientViT_B0.__doc__
EfficientViT_B3.__doc__ = EfficientViT_B0.__doc__
EfficientViT_L0.__doc__ = EfficientViT_B0.__doc__
EfficientViT_L1.__doc__ = EfficientViT_B0.__doc__
EfficientViT_L2.__doc__ = EfficientViT_B0.__doc__
EfficientViT_L3.__doc__ = EfficientViT_B0.__doc__

cascaded_mhsa_with_multi_head_position.__doc__ = __m_head_doc__ + """
Cascaded multi head self attention with MultiHeadPositionalEmbedding. Defined as function, not layer.
Cascaded calling flow performing multi head attention. Also using `Conv2D + BatchNorm` for `query` / `key` / `value` / `output`,
and an additional `DepthwiseConv2D` on `query` with `kernel_size`.

input: `[batch, height, width, channel]` if `channels_last`, or `[batch, channel, height, width]` if `channels_first`.
output: `[batch, height, width, channel]` if `channels_last`, or `[batch, channel, height, width]` if `channels_first`.

Args:
  inputs: input tensor.
  num_heads: Number of attention heads.
  key_dim: Size of each attention head for query, key. Default `0` for `key_dim = inputs.shape[-1] // num_heads`.
  kernel_sizes: int or list value indicates kernel_size for each head. Length must larger than max of num_heads.
  activation: activation used for output, default `relu`.

Examples:

>>> from keras_cv_attention_models import attention_layers
>>> inputs = keras.layers.Input([14, 16, 256])
>>> nn = attention_layers.cascaded_mhsa_with_multi_head_position(inputs, num_heads=1, name="attn_")
>>> print(f"{nn.shape = }")
# nn.shape = TensorShape([None, 14, 16, 256])

>>> mm = keras.models.Model(inputs, nn)
>>> mm.summary()
>>> print({ii.name: ii.shape for ii in mm.weights})
# {'attn_1_qkv_conv/kernel:0': TensorShape([1, 1, 256, 768]),
#  'attn_1_qkv_bn/gamma:0': TensorShape([768]),
#  'attn_1_qkv_bn/beta:0': TensorShape([768]),
#  'attn_1_qkv_bn/moving_mean:0': TensorShape([768]),
#  'attn_1_qkv_bn/moving_variance:0': TensorShape([768]),
#  'attn_1_query_dw_conv/depthwise_kernel:0': TensorShape([5, 5, 256, 1]),
#  'attn_1_query_bn/gamma:0': TensorShape([256]),
#  'attn_1_query_bn/beta:0': TensorShape([256]),
#  'attn_1_query_bn/moving_mean:0': TensorShape([256]),
#  'attn_1_query_bn/moving_variance:0': TensorShape([256]),
#  'attn_1_attn_pos/positional_embedding:0': TensorShape([224]),
#  'attn_outconv/kernel:0': TensorShape([1, 1, 256, 256]),
#  'attn_out_bn/gamma:0': TensorShape([256]),
#  'attn_out_bn/beta:0': TensorShape([256]),
#  'attn_out_bn/moving_mean:0': TensorShape([256]),
#  'attn_out_bn/moving_variance:0': TensorShape([256])}
"""



import numpy as np
import tensorflow as tf
from tqdm import tqdm
from keras_cv_attention_models.plot_func import get_plot_cols_rows, stack_and_plot_images

""" visualize_filters """


def __gradient_ascent_step__(feature_extractor, image_var, filter_index, optimizer):
    with tf.GradientTape() as tape:
        tape.watch(image_var)
        activation = feature_extractor(tf.expand_dims(image_var[0], 0))
        # We avoid border artifacts by only involving non-border pixels in the loss.
        # filter_activation = activation[:, 2:-2, 2:-2, filter_index]
        filter_activation = activation[..., filter_index]
        loss = tf.reduce_mean(filter_activation)
    # Compute gradients.
    grads = tape.gradient(loss, image_var)
    # Normalize gradients.
    grads = tf.math.l2_normalize(grads)
    # image_var += 10.0 * grads # 10.0 is learning_rate
    optimizer.apply_gradients(zip(grads * -1, image_var))  # For SGD, w = w - learning_rate * g
    return loss, image_var


def __initialize_image__(img_width, img_height, rescale_mode="tf", value_range=0.125):
    # We start from a gray image with some random noise. Pick image vvalue range in middle
    min_val, max_val = int(128 - 128 * value_range), int(128 + 128 * value_range)
    img = tf.random.uniform((img_width, img_height, 3), min_val, max_val)
    # For ResNet50V2 expects inputs in the range [-1, +1], results in `[-0.25, 0.25]`
    return tf.keras.applications.imagenet_utils.preprocess_input(img, mode=rescale_mode).numpy()


def __deprocess_image__(img, crop_border=0.1):
    # Normalize array: center on 0., ensure variance is 0.15
    img -= img.mean(axis=(0, 1))
    img /= img.std(axis=(0, 1)) + 1e-5
    img *= 0.15

    # Center crop
    hh_crop, ww_crop = int(img.shape[0] * crop_border), int(img.shape[1] * crop_border)
    img = img[hh_crop:-hh_crop, ww_crop:-ww_crop]

    # Clip to [0, 1]
    img += 0.5
    img = np.clip(img, 0, 1)

    # Convert to RGB array
    img *= 255
    img = np.clip(img, 0, 255).astype("uint8")
    return img


def visualize_filters(
    model,
    layer_name="auto",
    filter_index_list=[0],
    input_shape=None,
    rescale_mode="auto",
    iterations=30,
    optimizer="SGD",  # "SGD" / "RMSprop" / "Adam"
    learning_rate="auto",
    value_range=0.125,
    random_magnitude=1.0,  # basic random value for `tf.roll` and `random_rotation` is `4` and `1`.
    crop_border=0.1,
    base_size=3,
):
    from tensorflow.keras.preprocessing.image import random_rotation

    # Set up a model that returns the activation values for our target layer
    # model = tf.keras.models.clone_model(model)
    if layer_name == "auto":
        layer = model.layers[-1]
        layer_name = layer.name
    else:
        layer = model.get_layer(name=layer_name)
    feature_extractor = tf.keras.Model(inputs=model.inputs[0], outputs=layer.output)
    input_shape = model.input_shape[1:-1] if input_shape is None else input_shape[:2]
    assert input_shape[0] is not None and input_shape[1] is not None
    print(">>>> Total filters for layer {}: {}, input_shape: {}".format(layer_name, layer.output_shape[-1], input_shape))

    auto_lr = {"ada": 0.1, "rms": 1.0, "sgd": 10.0}
    if isinstance(optimizer, str):
        # After TF 2.11.0, seems we cannot call `apply_gradients` on `image_var`, use legacy instead...
        tf_optimizers = tf.optimizers.legacy if hasattr(tf.optimizers, "legacy") else tf.optimizers

        optimizer = optimizer.lower()
        learning_rate = auto_lr.get(optimizer[:3], 1.0) if learning_rate == "auto" else learning_rate
        if optimizer.startswith("rms"):
            optimizer = tf_optimizers.RMSprop(learning_rate, rho=0.999)
        elif optimizer == "adam":
            optimizer = tf_optimizers.Adam(learning_rate)
        elif optimizer == "sgd":
            optimizer = tf_optimizers.SGD(learning_rate)

    if rescale_mode.lower() == "auto":
        rescale_mode = getattr(model, "rescale_mode", "torch")
        print(">>>> rescale_mode:", rescale_mode)

    # We run gradient ascent for [iterations] steps
    losses, filter_images = [], []
    for filter_index in filter_index_list:
        image = __initialize_image__(input_shape[0], input_shape[1], rescale_mode, value_range=value_range)
        for iteration in tqdm(range(iterations), desc="Processing filter %d" % (filter_index,)):
            image = random_rotation(image, 1 * random_magnitude, row_axis=0, col_axis=1, channel_axis=2, fill_mode="reflect")
            image = tf.roll(image, shift=np.random.randint(-4 * random_magnitude, 4 * random_magnitude, size=2), axis=[0, 1])
            image_var = [tf.Variable(image)]
            # optimizer.build(image_var + feature_extractor.trainable_variables)
            loss, image_var = __gradient_ascent_step__(feature_extractor, image_var, filter_index, optimizer)
            image = image_var[0].numpy()
        # Decode the resulting input image
        image = __deprocess_image__(image, crop_border)
        losses.append(loss.numpy())
        filter_images.append(image)

    ax, _ = stack_and_plot_images(filter_images, base_size=base_size)
    return losses, np.stack(filter_images), ax



import torch


""" Functions """


def constant(value=0):
    return Constant(value=value)


def glorot_normal():
    return GlorotNormal()


def glorot_uniform():
    return GlorotUniform()


def he_normal():
    return HeNormal()


def he_uniform():
    return HeUniform()


def ones():
    return Ones()


def random_normal(mean=0.0, stddev=1e-6, seed=None):
    return RandomNormal(mean=mean, stddev=stddev, seed=seed)


def random_uniform(minval=-0.05, maxval=0.05, seed=None):
    return RandomUniform(minval=minval, maxval=maxval, seed=seed)


def truncated_normal(mean=0.0, stddev=1e-6, seed=None):
    return TruncatedNormal(mean=mean, stddev=stddev, seed=seed)


def zeros():
    return Zeros()


""" Classes """


class Initializer:
    def __init__(self, seed=None):
        self.seed = seed

    @classmethod
    def from_config(cls, config):
        config.pop("dtype", None)
        return cls(**config)


class Constant(Initializer):
    def __init__(self, value=0):
        self.value = value
        super().__init__(seed=None)

    def __call__(self, shape, dtype=None, **kwargs):
        if hasattr(self.value, "shape") and tuple(self.value.shape) == tuple(shape):
            return self.value
        else:
            return torch.nn.init.constant_(torch.empty(shape), val=self.value)

    def get_config(self):
        return {"value": self.value}


class GlorotNormal(Initializer):
    def __call__(self, shape, dtype=None, **kwargs):
        return torch.nn.init.xavier_normal_(torch.empty(shape))


class GlorotUniform(Initializer):
    def __call__(self, shape, dtype=None, **kwargs):
        return torch.nn.init.xavier_uniform_(torch.empty(shape))


class HeNormal(Initializer):
    def __call__(self, shape, dtype=None, **kwargs):
        return torch.nn.init.kaiming_normal_(torch.empty(shape))


class HeUniform(Initializer):
    def __call__(self, shape, dtype=None, **kwargs):
        return torch.nn.init.kaiming_uniform_(torch.empty(shape))


class Ones(Initializer):
    def __call__(self, shape, dtype=None, **kwargs):
        return torch.nn.init.ones_(torch.empty(shape))


class RandomNormal(Initializer):
    def __init__(self, mean=0.0, stddev=0.05, seed=None):
        self.mean, self.stddev = mean, stddev
        super().__init__(seed=seed)

    def __call__(self, shape, dtype=None, **kwargs):
        return torch.nn.init.normal_(torch.empty(shape), mean=self.mean, std=self.stddev)

    def get_config(self):
        return {"mean": self.mean, "stddev": self.stddev}


class RandomUniform(Initializer):
    def __init__(self, minval=-0.05, maxval=0.05, seed=None):
        self.minval, self.maxval = minval, maxval
        super().__init__(seed=seed)

    def __call__(self, shape, dtype=None, **kwargs):
        return torch.nn.init.uniform_(torch.empty(shape), a=self.minval, b=self.maxval)

    def get_config(self):
        return {"minval": self.minval, "maxval": self.maxval}


class TruncatedNormal(Initializer):
    def __init__(self, mean=0.0, stddev=0.05, seed=None):
        self.mean, self.stddev = mean, stddev
        super().__init__(seed=seed)

    def __call__(self, shape, dtype=None, **kwargs):
        return torch.nn.init.trunc_normal_(torch.empty(shape), mean=self.mean, std=self.stddev)

    def get_config(self):
        return {"mean": self.mean, "stddev": self.stddev}


class VarianceScaling(Initializer):
    def __init__(self, scale=1.0, mode="fan_in", distribution="truncated_normal", seed=None):
        # scale=2.0, mode="fan_in", distribution="uniform", seed=seed  # HeUniform
        # scale=2.0, mode="fan_in", distribution="truncated_normal", seed=seed  # HeNormal
        # scale=1.0, mode="fan_in", distribution="uniform", seed=seed  # LecunUniform
        # scale=1.0, mode="fan_in", distribution="truncated_normal", seed=seed  # LecunNormal
        # scale=1.0, mode="fan_avg", distribution="uniform", seed=seed  # GlorotUniform
        # scale=1.0, mode="fan_avg", distribution="truncated_normal", seed=seed  # GlorotNormal
        self.scale, self.mode, self.distribution, self.seed = scale, mode, distribution, seed

    def __call__(self, shape, dtype=None, **kwargs):
        return torch.zeros(shape)  # [TODO]

    def get_config(self):
        return {"scale": self.scale, "mode": self.mode, "distribution": self.distribution}


class Zeros(Initializer):
    def __call__(self, shape, dtype=None, **kwargs):
        return torch.nn.init.zeros_(torch.empty(shape))



"""
Copied from ultralytics/yolo/engine/validator.py. Original one will fuse bn if use directly on model
"""
import copy
import torch
from tqdm import tqdm
from ultralytics.models import yolo
from ultralytics.utils.ops import Profile
from ultralytics.data.utils import check_det_dataset
from pathlib import Path


class Validator:
    def __init__(self, model, val_loader, save_dir="./test", cfg={}):
        cfg = copy.copy(cfg)
        cfg.rect, cfg.mode = True, "val"
        self.device = next(model.parameters()).device
        cfg.half = self.device.type == "cuda"

        validator = yolo.detect.DetectionValidator(val_loader, save_dir=Path(save_dir), args=cfg.__dict__)
        # def eval(validator, model):
        validator.data = check_det_dataset(validator.args.data)
        validator.device = self.device
        self.n_batches = len(validator.dataloader)
        self.validator, self.model = validator, model

    def __call__(self):
        is_pre_training = self.model.training
        if self.device.type == "cuda":
            self.model.half()
        self.model.eval()
        dt = Profile(), Profile(), Profile()
        desc = self.validator.get_desc()

        bar = tqdm(self.validator.dataloader, desc, self.n_batches, bar_format="{l_bar}{bar:10}{r_bar}")
        self.validator.init_metrics(self.model)
        self.validator.jdict = []  # empty before each val
        with torch.no_grad():
            for batch_i, batch in enumerate(bar):
                self.validator.run_callbacks("on_val_batch_start")
                self.validator.batch_i = batch_i
                # preprocess
                with dt[0]:
                    batch = self.validator.preprocess(batch)

                # inference
                with dt[1]:
                    preds = self.model(batch["img"])

                # postprocess
                with dt[2]:
                    preds = self.validator.postprocess(preds)

                self.validator.update_metrics(preds, batch)
                if self.validator.args.plots and batch_i < 3:
                    self.validator.plot_val_samples(batch, batch_i)
                    self.validator.plot_predictions(batch, preds, batch_i)

            # self.validator.run_callbacks('on_val_batch_end')
        stats = self.validator.get_stats()
        self.validator.check_stats(stats)
        self.validator.print_results()
        # validator.speed = dict(zip(validator.speed.keys(), (x.t / len(validator.dataloader.dataset) * 1E3 for x in dt)))
        self.validator.finalize_metrics()
        # self.validator.run_callbacks('on_val_end')
        stats = self.validator.eval_json(stats)  # update stats
        if is_pre_training:
            self.model.train()
        if self.device.type == "cuda":
            self.model.float()
        return stats


if __name__ == "__main__":
    os.environ["KECAM_BACKEND"] = "torch"
    sys.path.append("../ultralytics/")
    from keras_cv_attention_models.yolov8 import yolov8, torch_wrapper, train, eval, data

    # from ultralytics import YOLO

    dataset_path = "coco128.yaml"
    train_loader, val_loader = data.get_data_loader(dataset_path=dataset_path, rect_val=True)
    cfg = train.FakeArgs(data=dataset_path, imgsz=640, iou=0.7, conf=0.001, single_cls=False, max_det=300, task="detect", mode="train", split="val", half=False)
    cfg.update(degrees=0.0, translate=0.1, scale=0.5, shear=0.0, perspective=0.0, hsv_h=0.015, hsv_s=0.7, hsv_v=0.4, flipud=0.0, fliplr=0.5)
    cfg.update(mask_ratio=4, overlap_mask=True, project=None, name=None, save_txt=False, save_hybrid=False, save_json=False, plots=False, verbose=True)

    # model = YOLO('../ultralytics/ultralytics/models/v8/yolov8n.yaml').model
    model = yolov8.YOLOV8_N(input_shape=(3, None, None), classifier_activation=None, pretrained=None)
    model = torch_wrapper.Detect(model)
    ee = eval.Validator(model, val_loader, cfg=cfg)
    ee()



from keras_cv_attention_models.pytorch_backend import functional

BUILDIN_METRICS = {}


def register_metrics(name=None):
    def decorator(arg):
        registered_names = name or [arg.__name__]
        registered_names = registered_names if isinstance(registered_names, (list, tuple)) else [registered_names]
        for registered_name in registered_names:
            if registered_name in BUILDIN_METRICS:
                raise ValueError(f"{registered_name} has already been registered to " f"{BUILDIN_METRICS[registered_name]}")
            BUILDIN_METRICS[registered_name] = arg
        return arg

    return decorator


class Metric:
    def __init__(self, name=None, **kwargs):
        super().__init__()
        self.name = name
        self.eval_only = False
        self.reset_state()

    def reset_state(self):
        pass

    def update_state(self, y_true, y_pred, sample_weight=None):
        pass

    def result(self):
        pass


@register_metrics(name=["acc", "accuracy"])
class Accuracy(Metric):
    def __init__(self, name="acc"):
        super().__init__(name=name)

    def reset_state(self):
        self.sum_value, self.passed_steps = 0.0, 0

    def update_state(self, y_true, y_pred, sample_weight=None):
        y_pred = functional.argmax(y_pred, axis=-1)
        if len(y_true.shape) > len(y_pred.shape):
            y_true = functional.argmax(y_true, axis=-1)
        cur_acc = functional.reduce_mean(functional.cast(y_true == y_pred, "float32"))
        self.sum_value = self.sum_value + cur_acc
        self.passed_steps += 1

    def result(self):
        return self.sum_value / self.passed_steps


@register_metrics(name=["acc5", "accuracy5"])
class Accuracy5(Metric):
    def __init__(self, name="acc5"):
        super().__init__(name=name)
        self.eval_only = True

    def reset_state(self):
        self.sum_value, self.passed_steps = 0.0, 0

    def update_state(self, y_true, y_pred, sample_weight=None):
        y_pred = functional.argsort(y_pred, direction="DESCENDING", axis=-1)[:, :5]
        if len(y_true.shape) >= len(y_pred.shape):
            y_true = functional.argmax(y_true, axis=-1)
        cur_acc = functional.reduce_mean(functional.convert_to_tensor([y_true[id] in y_pred[id] for id in range(y_true.shape[0])], "float32"))
        self.sum_value = self.sum_value + cur_acc
        self.passed_steps += 1

    def result(self):
        return self.sum_value / self.passed_steps



from keras_cv_attention_models.yolov8.yolov8 import (
    YOLOV8Backbone,
    YOLOV8,
    YOLOV8_N,
    YOLOV8_S,
    YOLOV8_M,
    YOLOV8_L,
    YOLOV8_X,
    YOLOV8_X6,
    YOLOV8_N_CLS,
    YOLOV8_S_CLS,
    YOLOV8_M_CLS,
    YOLOV8_L_CLS,
    YOLOV8_X_CLS,
    switch_to_deploy,
)
from keras_cv_attention_models.yolov8.yolo_nas import YOLO_NAS, YOLO_NAS_S, YOLO_NAS_M, YOLO_NAS_L

__head_doc__ = """
Keras implementation of [Github ultralytics/ultralytics](https://github.com/ultralytics/ultralytics).
"""

__detector_head_doc__ = """
Args:
  backbone: backbone model, could be any model with pyramid stage structure.
      Default None for YOLOV8Backbone.
"""

__detector_tail_doc__ = """  features_pick: specific `layer names` or `pyramid feature indexes` from backbone model.
        Default `[-3, -2, -1]` means using the last 3 pyramid feature output from backbone.
  regression_len: bbox output len, typical value is 4, for yolov8 reg_max=16 -> regression_len = 16 * 4 == 64.
      For YOLO_NAS, it's 17 * 4 == 68.
  use_reparam_conv: boolean value if using `reparam_conv_bn` instead of `conv_bn` block in all `csp_block`s.
      Roughly, `reparam_conv_bn` is a block with `Conv_3x3+BN(inputs) + Conv_1x1(inputs) + inputs`,
      and can be reparametered to a single `Conv_3x3` layer after training.
  paf_parallel_mode: False for YOLOV8_X6 and True for others.
      If False, only concat `short` and the last `deep` one in `path_aggregation_fpn` module.
  anchors_mode: one of ["efficientdet", "anchor_free", "yolor", "yolov8"], controls which anchor to use.
      - efficientdet anchors default settings: use_object_scores=False, num_anchors=9, anchor_scale=4,
          aspect_ratios=[1, 2, 0.5], num_scales=3, grid_zero_start=False.
      - anchor_free default settings: use_object_scores=True, num_anchors=1, anchor_scale=1,
          aspect_ratios=[1], num_scales=1, grid_zero_start=True.
      - yolor default settings: use_object_scores=True, num_anchors=3.
      - yolov8 default settings: use_object_scores=False, num_anchors=1, anchor_scale=1,
          aspect_ratios=[1], num_scales=1, grid_zero_start=False.
      Default "yolov8".
  num_anchors: number of anchors for a single grid point, should be same with dataset used value.
      Default "auto" means: anchors_mode=="anchor_free" / "yolov8" -> 1, anchors_mode=="yolor" -> 3, else 9.
  use_object_scores: bollean value if model header output includes `object_scores`.
      Default "auto" means: True if anchors_mode=="anchor_free" or anchors_mode=="yolor", else False.
  input_shape: input shape if backbone is None, else will use input_shape from backbone.
  num_classes: total output classes. Set `0` to disable `classifier` output. Default 80 for COCO.
  activation: activation used in whole model, default `swish`. Default "swish".
  classifier_activation: The activation function to use for classifier output if `num_classes > 0`.
      Set `classifier_activation=None` to return the logits of the "top" layer. Default `sigmoid`.
  freeze_backbone: set `True` for `backbone.trainable = False`. Default `False`.
  pretrained: one of `None` (random initialization) or 'coco' (pre-training on COCO).
      Will try to download and load pre-trained model weights if not None. Default `coco`.
  pyramid_levels_min: anchors inititial parameter for model prediction, not affecting model architecture. Default `3`.
      pyramid_levels_max is calculated as `pyramid_levels_min + len(features_pick) - 1`.
  anchor_scale: anchors inititial parameter for model prediction, not affecting model architecture.
      Default "auto" means: 1 if (anchors_mode=="anchor_free" or anchors_mode=="yolor"), else 4.
  rescale_mode: model precessing input, not for model structure. Defulat "raw01" means input value in range `[0, 1]`.

Returns:
    A `keras.Model` instance.
"""

YOLOV8.__doc__ = __head_doc__ + __detector_head_doc__ + """
[YOLOV8Backbone parameters]
  csp_channels: YOLOV8Backbone backbone channel for each block, default: `[32, 64, 128, 256]`.
  csp_depthes: YOLOV8Backbone backbone depth for each block, default: `[1, 2, 2, 1]`.

[YOLOV8 parameters]
  model_name: string, model name.

[Detector parameters]
""" + __detector_tail_doc__ + """
Model architectures:
  | Model     | Params | FLOPs  | Input | COCO val AP | test AP |
  | --------- | ------ | ------ | ----- | ----------- | ------- |
  | YOLOV8_N  | 3.16M  | 4.39G  | 640   | 37.3        |         |
  | YOLOV8_S  | 11.17M | 14.33G | 640   | 44.9        |         |
  | YOLOV8_M  | 25.90M | 39.52G | 640   | 50.2        |         |
  | YOLOV8_L  | 43.69M | 82.65G | 640   | 52.9        |         |
  | YOLOV8_X  | 68.23M | 129.0G | 640   | 53.9        |         |
  | YOLOV8_X6 | 97.42M | 522.6G | 1280  | 56.7 ?      |         |

  | Model                    | Params | FLOPs  | Input | COCO val AP | test AP |
  | ------------------------ | ------ | ------ | ----- | ----------- | ------- |
  | YOLO_NAS_S               | 12.88M | 16.96G | 640   | 47.5        |         |
  | - use_reparam_conv=False | 12.18M | 15.92G | 640   | 47.5        |         |
  | YOLO_NAS_M               | 33.86M | 47.12G | 640   | 51.55       |         |
  | - use_reparam_conv=False | 31.92M | 43.91G | 640   | 51.55       |         |
  | YOLO_NAS_L               | 44.53M | 64.53G | 640   | 52.22       |         |
  | - use_reparam_conv=False | 42.02M | 59.95G | 640   | 52.22       |         |
"""

YOLOV8_N.__doc__ = __head_doc__ + __detector_head_doc__ + __detector_tail_doc__
YOLOV8_S.__doc__ = YOLOV8_N.__doc__
YOLOV8_M.__doc__ = YOLOV8_N.__doc__
YOLOV8_L.__doc__ = YOLOV8_N.__doc__
YOLOV8_X.__doc__ = YOLOV8_N.__doc__
YOLOV8_X6.__doc__ = YOLOV8_N.__doc__

YOLO_NAS.__doc__ = YOLOV8.__doc__
YOLO_NAS_S.__doc__ = YOLOV8_N.__doc__
YOLO_NAS_M.__doc__ = YOLOV8_N.__doc__
YOLO_NAS_L.__doc__ = YOLOV8_N.__doc__

__classifier_tail_doc__ = """  input_shape: it should have exactly 3 inputs channels, like `(224, 224, 3)`.
  num_classes: number of classes to classify images into. Set `0` to exclude top layers.
  activation: activation used in whole stacks, default `gelu`.
  dropout: top dropout rate if top layers is included. Default 0.
  classifier_activation: A `str` or callable. The activation function to use on the "top" layer if `num_classes > 0`.
      Set `classifier_activation=None` to return the logits of the "top" layer.
      Default is `None`.
  pretrained: one of `None` (random initialization) or 'imagenet' (pre-training on ImageNet).
      Will try to download and load pre-trained model weights if not None.
  **kwargs: other parameters if available.
Returns:
    A `keras.Model` instance.
"""

YOLOV8Backbone.__doc__ = __head_doc__ + """
Args:
  channels: channel for each block, default: `[32, 64, 128, 256]`.
  depthes: depth for each block, default: `[1, 2, 2, 1]`.
""" + __classifier_tail_doc__ + """
Model architectures:
  | Model        | Params | FLOPs@640 | FLOPs@224 | Input | Top1 Acc |
  | ------------ | ------ | --------- | --------- | ----- | -------- |
  | YOLOV8_N_CLS | 2.72M  | 1.65G     | 203.7M    | 224   | 66.6     |
  | YOLOV8_S_CLS | 6.36M  | 6.24G     | 765.7M    | 224   | 72.3     |
  | YOLOV8_M_CLS | 17.05M | 20.85G    | 2.56G     | 224   | 76.4     |
  | YOLOV8_L_CLS | 37.48M | 49.41G    | 6.05G     | 224   | 78.0     |
  | YOLOV8_X_CLS | 57.42M | 76.96G    | 9.43G     | 224   | 78.4     |
"""

YOLOV8_N_CLS.__doc__ = __head_doc__ + """
Args:
""" + __classifier_tail_doc__

YOLOV8_S_CLS.__doc__ = YOLOV8_N_CLS.__doc__
YOLOV8_M_CLS.__doc__ = YOLOV8_N_CLS.__doc__
YOLOV8_L_CLS.__doc__ = YOLOV8_N_CLS.__doc__
YOLOV8_X_CLS.__doc__ = YOLOV8_N_CLS.__doc__



import math
import numpy as np
from keras_cv_attention_models import backend
from keras_cv_attention_models.backend import layers, functional, models, image_data_format
from keras_cv_attention_models.models import register_model
from keras_cv_attention_models.attention_layers import (
    ChannelAffine,
    MlpPairwisePositionalEmbedding,
    conv2d_no_bias,
    depthwise_conv2d_no_bias,
    batchnorm_with_activation,
    add_with_layer_scale_and_drop_block,
    pad_to_divisible_by_window_size,
    reverse_padded_for_window_size,
    window_reverse,
    multi_head_self_attention,
    layer_norm,
    mlp_block,
    add_pre_post_process,
)
from keras_cv_attention_models.download_and_load import reload_model_weights

BATCH_NORM_EPSILON = 1e-4
LAYER_NORM_EPSILON = 1e-6

PRETRAINED_DICT = {
    "fastervit_0": {"imagenet": {224: "afafa8c70005e14f24f7b544fd4f0caf"}},
    "fastervit_1": {"imagenet": {224: "3b309f8c894a8f822e3a0cc374381b93"}},
    "fastervit_2": {"imagenet": {224: "b1f7ff0d70859199c24af59716adec2a"}},
    "fastervit_3": {"imagenet": {224: "dea02b352abbb2e76ceb241853b55c56"}},
    "fastervit_4": {"imagenet": {224: "9cd2a9df1de436c37da92a164d333a78"}},
    "fastervit_5": {"imagenet": {224: "4448add6981ec321ba7d6bd780886281"}},
    "fastervit_6": {"imagenet": {224: ["342756534860f0602ab2715445ce9c20", "00174cfde85f89a17cf74d4384191ea1"]}},
}


def res_conv_bn_block(inputs, layer_scale=0, drop_rate=0, activation="gelu", name=""):
    input_channel = inputs.shape[-1 if image_data_format() == "channels_last" else 1]

    nn = conv2d_no_bias(inputs, input_channel, kernel_size=3, use_bias=True, padding="same", name=name + "1_")  # epsilon=1e-5
    nn = batchnorm_with_activation(nn, activation=activation, name=name + "1_")

    nn = conv2d_no_bias(nn, input_channel, kernel_size=3, use_bias=True, padding="same", name=name + "2_")
    nn = batchnorm_with_activation(nn, activation=None, name=name + "2_")  # epsilon=1e-5
    nn = add_with_layer_scale_and_drop_block(inputs, nn, layer_scale=layer_scale, drop_rate=drop_rate, name=name)
    return nn


def attention_mlp_block(
    inputs, carrier_tokens=None, num_heads=4, attn_height=-1, mlp_ratio=4, pos_scale=-1, layer_scale=0, drop_rate=0, activation="gelu", name=""
):
    input_channel = inputs.shape[-1]
    if len(inputs.shape) == 4:
        attn_height, attn_width = inputs.shape[1:-1]
    else:
        attn_height = attn_height if attn_height > 0 else int(inputs.shape[1] ** 0.5)  # inputs is square shape, attn_height == attn_width
        attn_width = inputs.shape[1] // attn_height

    if carrier_tokens is not None:
        num_carrier_tokens = int(np.prod(carrier_tokens.shape[1:-1]))
        inputs = functional.concat([carrier_tokens, inputs], axis=1)
    # print(f"{attn_height = }, {attn_width = }")

    """ attention """
    attn = layer_norm(inputs, epsilon=LAYER_NORM_EPSILON, axis=-1, name=name + "attn_")
    pos_emb = MlpPairwisePositionalEmbedding(attn_height=attn_height, attn_width=attn_width, pos_scale=pos_scale, name=name + "attn_pos")
    attn = multi_head_self_attention(attn, num_heads=num_heads, pos_emb=pos_emb, qkv_bias=True, out_bias=True, name=name + "attn_")
    if layer_scale > 0:
        # Requries for `do_propagation=True`
        ct_gamma_layer = ChannelAffine(use_bias=False, weight_init_value=layer_scale, axis=-1, name=name + "ct_gamma")
        attn = ct_gamma_layer(attn)
    else:
        ct_gamma_layer = None
    attn_out = add_with_layer_scale_and_drop_block(inputs, attn, layer_scale=0, drop_rate=drop_rate, axis=-1, name=name + "attn_out_")

    """ MLP """
    nn = layer_norm(attn_out, epsilon=LAYER_NORM_EPSILON, axis=-1, name=name + "mlp_")
    nn = mlp_block(nn, input_channel * mlp_ratio, use_bias=True, activation=activation, name=name + "mlp_")
    nn = add_with_layer_scale_and_drop_block(attn_out, nn, layer_scale=layer_scale, drop_rate=drop_rate, axis=-1, name=name + "mlp_")
    # print(f"{nn.shape = }, {attn_height = }, {attn_width = }")
    if carrier_tokens is not None:
        carrier_tokens, nn = functional.split(nn, [num_carrier_tokens, attn_height * attn_width], axis=1)
    return nn, carrier_tokens, ct_gamma_layer


def hierarchical_attention(
    inputs,
    num_heads=4,
    attn_height=-1,
    mlp_ratio=4,
    carrier_tokens=None,
    token_size=2,
    use_propagation=False,
    pos_scale=-1,
    layer_scale=0,
    drop_rate=0,
    activation="gelu",
    name="",
):
    attn_kwargs = {"num_heads": num_heads, "mlp_ratio": mlp_ratio, "pos_scale": pos_scale, "layer_scale": layer_scale, "drop_rate": drop_rate}
    # print(f"{inputs.shape = }, {attn_kwargs = }")
    if carrier_tokens is not None:
        # print(f"{carrier_tokens.shape = }")
        ct_channels = carrier_tokens.shape[-1]
        ct_patch_height, ct_patch_width = carrier_tokens.shape[1] // token_size, carrier_tokens.shape[2] // token_size
        carrier_tokens = MlpPairwisePositionalEmbedding(pos_scale=pos_scale, use_absolute_pos=True, name=name + "hat_pos")(carrier_tokens)
        carrier_tokens = functional.reshape(carrier_tokens, [-1, int(np.prod(carrier_tokens.shape[1:-1])), ct_channels])
        carrier_tokens, _, ct_gamma_layer = attention_mlp_block(carrier_tokens, **attn_kwargs, activation=activation, name=name + "hat_")
        # print(f"{carrier_tokens.shape = }")
        carrier_tokens = functional.reshape(carrier_tokens, [-1, token_size, ct_patch_width, token_size * ct_channels])
        carrier_tokens = functional.transpose(carrier_tokens, [0, 2, 1, 3])  # [batch * patch_height, patch_width, window_height, window_width * channel]
        carrier_tokens = functional.reshape(carrier_tokens, [-1, token_size * token_size, ct_channels])
    else:
        ct_gamma_layer = None

    pre = MlpPairwisePositionalEmbedding(pos_scale=pos_scale, attn_height=attn_height, use_absolute_pos=True, name=name + "pre_attn_pos")(inputs)
    # Take carrier_tokens in, for haddling shape there
    nn, carrier_tokens, _ = attention_mlp_block(pre, carrier_tokens, attn_height=attn_height, **attn_kwargs, activation=activation, name=name)

    if use_propagation and carrier_tokens is not None:
        nn = do_propagation(nn, carrier_tokens, ct_gamma_layer, token_size)
    elif carrier_tokens is not None:
        carrier_tokens = window_reverse(
            carrier_tokens, patch_height=ct_patch_height, patch_width=ct_patch_width, window_height=token_size, window_width=token_size
        )
    return nn, carrier_tokens


def do_propagation(inputs, carrier_tokens, ct_gamma_layer=None, token_size=2):
    height = width = int(float(inputs.shape[1]) ** 0.5)  # height, width should be all equal to window_size
    carrier_tokens = functional.reshape(carrier_tokens, [-1, token_size, token_size, carrier_tokens.shape[-1]])
    carrier_tokens = carrier_tokens if image_data_format() == "channels_last" else layers.Permute([3, 1, 2])(carrier_tokens)
    carrier_tokens = functional.resize(carrier_tokens, size=(height, width), method="nearest")
    carrier_tokens = carrier_tokens if image_data_format() == "channels_last" else layers.Permute([2, 3, 1])(carrier_tokens)
    carrier_tokens = functional.reshape(carrier_tokens, [-1, height * width, carrier_tokens.shape[-1]])
    carrier_tokens = ct_gamma_layer(carrier_tokens) if ct_gamma_layer is not None else carrier_tokens
    return inputs + carrier_tokens


def global_carrier_tokens(inputs, window_size=(7, 7), token_size=2, name=""):
    input_size = inputs.shape[1:-1] if image_data_format() == "channels_last" else inputs.shape[2:]
    outputs = [token_size * math.ceil(input_size[0] / window_size[0]), token_size * math.ceil(input_size[1] / window_size[1])]
    strides = [input_size[0] // outputs[0], input_size[1] // outputs[1]]
    pool_size = [input_size[0] - (outputs[0] - 1) * strides[0], input_size[1] - (outputs[1] - 1) * strides[1]]

    nn = depthwise_conv2d_no_bias(inputs, kernel_size=3, padding="same", use_bias=True, name=name)
    nn = layers.AvgPool2D(pool_size=pool_size, strides=strides)(nn)
    # print(f"[global_carrier_tokens] {inputs.shape = }, {outputs = }, {pool_size = }, {strides = }, {nn.shape = }")
    # nn = window_partition(nn, token_size, token_size)
    return nn


def FasterViT(
    num_blocks=[2, 3, 6, 5],
    num_heads=[2, 4, 8, 16],
    # window_sizes=[8, 8, 7, 7],
    window_ratios=[1, 1, 2, 1],
    stem_hidden_dim=64,
    embed_dim=64,
    mlp_ratio=4,
    carrier_token_size=2,
    pos_scale=-1,  # If pretrained weights are from different input_shape or window_size, pos_scale is previous actually using window_size
    use_propagation=False,
    # use_layernorm_output=False,
    input_shape=(224, 224, 3),
    num_classes=1000,
    activation="gelu",
    drop_connect_rate=0,
    layer_scale=0,
    classifier_activation="softmax",
    dropout=0,
    pretrained=None,
    model_name="fastervit",
    kwargs=None,
):
    """Patch stem"""
    # Regard input_shape as force using original shape if len(input_shape) == 4,
    # else assume channel dimension is the one with min value in input_shape, and put it first or last regarding image_data_format
    input_shape = backend.align_input_shape_by_image_data_format(input_shape)
    inputs = layers.Input(input_shape)
    nn = conv2d_no_bias(inputs, stem_hidden_dim, 3, strides=2, padding="same", name="stem_1_")
    nn = batchnorm_with_activation(nn, epsilon=BATCH_NORM_EPSILON, activation="relu", name="stem_1_")
    nn = conv2d_no_bias(nn, embed_dim, 3, strides=2, padding="same", name="stem_2_")
    nn = batchnorm_with_activation(nn, epsilon=BATCH_NORM_EPSILON, activation="relu", name="stem_2_")

    is_conv_blocks = [True, True, False, False]
    carrier_token_sizes = [0, 0, carrier_token_size, 0]

    """ stage [1, 2, 3, 4] """
    total_blocks = sum(num_blocks)
    global_block_id = 0
    block_args = zip(num_blocks, is_conv_blocks, window_ratios, num_heads, carrier_token_sizes)
    for stack_id, (num_block, is_conv_block, window_ratio, num_head, carrier_token_size) in enumerate(block_args):
        stack_name = "stack{}_".format(stack_id + 1)
        out_channels = embed_dim * (2**stack_id)

        if stack_id > 0:
            nn = layer_norm(nn, epsilon=LAYER_NORM_EPSILON, name=stack_name + "downsample_")
            nn = conv2d_no_bias(nn, out_channels, 3, strides=2, padding="same", name=stack_name + "downsample_")

        if not is_conv_block:
            height, width = nn.shape[1:-1] if image_data_format() == "channels_last" else nn.shape[2:]
            window_size = (math.ceil(height / window_ratio), math.ceil(width / window_ratio))
            # print(f"{window_size = }")
            # use_carrier_token = nn.shape[1] > window_size[1] or nn.shape[2] > window_size[2]
            if carrier_token_size > 0:
                carrier_tokens = global_carrier_tokens(nn, window_size, token_size=carrier_token_size, name=stack_name + "token_")
                carrier_tokens = carrier_tokens if image_data_format() == "channels_last" else layers.Permute([2, 3, 1])(carrier_tokens)
            else:
                carrier_tokens = None
            nn = nn if image_data_format() == "channels_last" else layers.Permute([2, 3, 1])(nn)

            nn, window_height, window_width, padding_height, padding_width = pad_to_divisible_by_window_size(nn, window_size)
            patch_height, patch_width = nn.shape[1] // window_height, nn.shape[2] // window_width
            # sr_ratio = max(patch_height, patch_width)
            if patch_height > 1 or patch_width > 1:
                nn = functional.reshape(nn, [-1, window_height, patch_width, window_width * out_channels])
                nn = functional.transpose(nn, [0, 2, 1, 3])  # [batch * patch_height, patch_width, window_height, window_width * channel]
                nn = functional.reshape(nn, [-1, window_height * window_width, out_channels])  # All using 3D, faster for ONNX inference [???]

        for block_id in range(num_block):
            name = "stack{}_block{}_".format(stack_id + 1, block_id + 1)
            block_drop_rate = drop_connect_rate * global_block_id / total_blocks
            cur_use_propagation = use_propagation if block_id == num_block - 1 else False
            global_block_id += 1
            if is_conv_block:
                nn = res_conv_bn_block(nn, layer_scale=0, drop_rate=block_drop_rate, activation=activation, name=name)
            else:
                # Save line width
                hat_kwargs = {"use_propagation": cur_use_propagation, "pos_scale": pos_scale, "layer_scale": layer_scale, "drop_rate": block_drop_rate}
                nn, carrier_tokens = hierarchical_attention(
                    nn, num_head, window_height, mlp_ratio, carrier_tokens, carrier_token_size, **hat_kwargs, activation=activation, name=name
                )

        if not is_conv_block:
            nn = window_reverse(nn, patch_height, patch_width, window_height, window_width)
            nn = reverse_padded_for_window_size(nn, padding_height, padding_width)
        # print(f"{nn.shape = }")
        nn = nn if is_conv_block or image_data_format() == "channels_last" else layers.Permute([3, 1, 2], name=stack_name + "permute_output")(nn)

    if num_classes > 0:
        nn = batchnorm_with_activation(nn, epsilon=BATCH_NORM_EPSILON, activation=None, name="pre_out_")
        nn = layers.GlobalAveragePooling2D(name="avg_pool")(nn)
        # nn = layers.LayerNormalization(axis=-1, epsilon=LAYER_NORM_EPSILON, name="post_ln")(nn) if use_layernorm_output else nn
        if dropout > 0:
            nn = layers.Dropout(dropout, name="head_drop")(nn)
        nn = layers.Dense(num_classes, dtype="float32", activation=classifier_activation, name="predictions")(nn)
    model = models.Model(inputs, nn, name=model_name)
    reload_model_weights(model, PRETRAINED_DICT, "fastervit", pretrained)

    add_pre_post_process(model, rescale_mode="torch")
    model.switch_to_deploy = lambda: switch_to_deploy(model)
    return model


def switch_to_deploy(model):
    from keras_cv_attention_models.model_surgery.model_surgery import convert_layers_to_deploy_inplace

    new_model = convert_layers_to_deploy_inplace(model)
    add_pre_post_process(new_model, rescale_mode=model.preprocess_input.rescale_mode, post_process=model.decode_predictions)
    return new_model


@register_model
def FasterViT0(input_shape=(224, 224, 3), num_classes=1000, classifier_activation="softmax", pretrained="imagenet", **kwargs):
    return FasterViT(**locals(), model_name="fastervit_0", **kwargs)


@register_model
def FasterViT1(input_shape=(224, 224, 3), num_classes=1000, classifier_activation="softmax", pretrained="imagenet", **kwargs):
    num_blocks = [1, 3, 8, 5]
    embed_dim = 80
    stem_hidden_dim = 32
    return FasterViT(**locals(), model_name="fastervit_1", **kwargs)


@register_model
def FasterViT2(input_shape=(224, 224, 3), num_classes=1000, classifier_activation="softmax", pretrained="imagenet", **kwargs):
    num_blocks = [3, 3, 8, 5]
    embed_dim = 96
    return FasterViT(**locals(), model_name="fastervit_2", **kwargs)


@register_model
def FasterViT3(input_shape=(224, 224, 3), num_classes=1000, classifier_activation="softmax", pretrained="imagenet", **kwargs):
    num_blocks = [3, 3, 12, 5]
    embed_dim = 128
    layer_scale = 1e-5
    use_propagation = True
    return FasterViT(**locals(), model_name="fastervit_3", **kwargs)


@register_model
def FasterViT4(input_shape=(224, 224, 3), num_classes=1000, classifier_activation="softmax", pretrained="imagenet", **kwargs):
    num_blocks = [3, 3, 12, 5]
    num_heads = [4, 8, 16, 32]
    embed_dim = 196
    layer_scale = 1e-5
    use_propagation = True
    return FasterViT(**locals(), model_name="fastervit_4", **kwargs)


@register_model
def FasterViT5(input_shape=(224, 224, 3), num_classes=1000, classifier_activation="softmax", pretrained="imagenet", **kwargs):
    num_blocks = [3, 3, 12, 5]
    num_heads = [4, 8, 16, 32]
    embed_dim = 320
    layer_scale = 1e-5
    use_propagation = True
    return FasterViT(**locals(), model_name="fastervit_5", **kwargs)


@register_model
def FasterViT6(input_shape=(224, 224, 3), num_classes=1000, classifier_activation="softmax", pretrained="imagenet", **kwargs):
    num_blocks = [3, 3, 16, 8]
    num_heads = [4, 8, 16, 32]
    embed_dim = 320
    layer_scale = 1e-5
    use_propagation = True
    return FasterViT(**locals(), model_name="fastervit_6", **kwargs)



import os
import numpy as np
from keras_cv_attention_models import backend
from keras_cv_attention_models.backend import layers

HDF5_OBJECT_HEADER_LIMIT = 64512


def reload_model_weights(
    model, pretrained_dict, sub_release, pretrained="imagenet", mismatch_class=None, force_reload_mismatch=False, request_resolution=-1, method=None
):
    if not isinstance(pretrained, (str, list, tuple)):
        return
    if isinstance(pretrained, (list, tuple)) or pretrained.endswith(".h5") or pretrained.endswith(".keras") or pretrained.endswith(".pt"):
        pretraineds = pretrained if isinstance(pretrained, (list, tuple)) else [pretrained]
        for pretrained in pretraineds:
            print(">>>> Load pretrained from:", pretrained)
            # model.load_weights(pretrained, by_name=True, skip_mismatch=True)
            load_weights_with_mismatch(model, pretrained, mismatch_class, request_resolution=request_resolution, method=method)
        return pretraineds

    file_hash = pretrained_dict.get(model.name, {}).get(pretrained, None)
    if file_hash is None:
        print(">>>> No pretrained available, model will be randomly initialized")
        return None

    if isinstance(file_hash, dict):
        # file_hash is a dict like {224: "aa", 384: "bb", 480: "cc"}
        if request_resolution == -1:
            input_height = model.input_shape[2]  # Either channels_last or channels_first, 2 aixs has to be a shape one
            if input_height is None:  # input_shape is (None, None, 3)
                request_resolution = max(file_hash.keys())
            else:
                request_resolution = min(file_hash.keys(), key=lambda ii: abs(ii - input_height))
        pretrained = "{}_".format(request_resolution) + pretrained
        file_hash = file_hash[request_resolution]
        # print(f"{request_resolution = }, {pretrained = }, {file_hash = }")
    elif request_resolution == -1:
        request_resolution = 224  # Default is 224

    if isinstance(file_hash, (list, tuple)):
        # For large weight file split in several pieces. In format aa.1.h5, aa.2.h5, ...
        pre_url = "https://github.com/leondgarse/keras_cv_attention_models/releases/download/{}/{}_{}.{}.h5"
        is_multi_files = True
    else:
        pre_url = "https://github.com/leondgarse/keras_cv_attention_models/releases/download/{}/{}_{}.h5"
        is_multi_files = False
        file_hash = [file_hash]

    pretrained_models = []
    for id, cur_file_hash in enumerate(file_hash):
        url = pre_url.format(sub_release, model.name, pretrained, id + 1) if is_multi_files else pre_url.format(sub_release, model.name, pretrained)
        file_name = os.path.basename(url)
        try:
            pretrained_model = backend.get_file(file_name, url, cache_subdir="models", file_hash=cur_file_hash)
        except:
            print("[Error] will not load weights, url not found or download failed:", url)
            return None
        else:
            print(">>>> Load pretrained from:", pretrained_model)
            # model.load_weights(pretrained_model, by_name=True, skip_mismatch=True)
            load_weights_with_mismatch(model, pretrained_model, mismatch_class, force_reload_mismatch, request_resolution, method)
            pretrained_models.append(pretrained_model)
    return pretrained_model[0] if len(pretrained_models) == 1 else pretrained_models


def load_weights_with_mismatch(model, weight_file, mismatch_class=None, force_reload_mismatch=False, request_resolution=-1, method=None):
    model.load_weights(weight_file, by_name=True, skip_mismatch=True)
    if len(model.input_shape) == 4 and backend.image_data_format() == "channels_first":
        input_height, input_width = model.input_shape[2], model.input_shape[3]
    elif len(model.input_shape) == 4:
        input_height, input_width = model.input_shape[1], model.input_shape[2]
    else:
        input_height, input_width = -1, -1

    # if mismatch_class is not None:
    if mismatch_class is not None and (force_reload_mismatch or request_resolution != input_height or request_resolution != input_width):
        try:
            import h5py

            print(">>>> Reload mismatched weights: {} -> {}".format(request_resolution, (input_height, input_width)))
            with h5py.File(weight_file, mode="r") as h5_file:
                weights = h5_file["model_weights"] if "model_weights" in h5_file else h5_file  # full model or weights only

                if isinstance(mismatch_class, (list, tuple)):
                    is_mismatch_class = lambda xx: any([isinstance(ii, mm) for mm in mismatch_class])
                else:
                    is_mismatch_class = lambda xx: isinstance(ii, mismatch_class)

                method_kwarg = {} if method is None else {"method": method}  # None for using class default one
                for ii in model.layers:
                    if is_mismatch_class(ii) and ii.name in weights:
                        print(">>>> Reload layer:", ii.name)
                        ss = weights[ii.name]
                        # ss = {ww.decode().split("/")[-1] : tf.convert_to_tensor(ss[ww]) for ww in ss.attrs['weight_names']}
                        ss = {ww.decode("utf8") if hasattr(ww, "decode") else ww: np.array(ss[ww]) for ww in ss.attrs["weight_names"]}
                        ss = {kk.split("/")[-1]: vv for kk, vv in ss.items()}
                        model.get_layer(ii.name).load_resized_weights(ss, **method_kwarg)
        except Exception as error:
            print("[Error] something went wrong in load_weights_with_mismatch:", error)
            pass


""" Save / load h5 weights from keras.saving.legacy.hdf5_format, supports both TF and PyTorch model, and convert_torch_weights_to_h5 """


class H5orKerasFileReader:
    """ Read `.h5` or `.keras` format weight file as a dict, `.keras` is the new saving format since TF 2.13
    >>> import kecam
    >>> mm = kecam.models.LLaMA2_15M()
    >>> mm.save('aa.keras')  # Or 'aa.h5'
    >>>
    >>> with kecam.download_and_load.H5orKerasFileReader('aa.keras') as ff:
    >>>     print({kk: [ii.shape for ii in vv] for kk, vv in ff.items()})
    >>>     kecam.download_and_load._load_layer_weights_nested_(mm, ff, debug=True)
    """
    def __init__(self, filepath):
        self.filepath = filepath
        self.is_keras_format = self.filepath.endswith(".keras")
        self.read = self.__read_keras_file__ if self.is_keras_format else self.__read_h5_file__

    def __enter__(self):
        return self.read()

    def __exit__(self, type, value, trace):
        self.close()

    def __read_h5_file__(self):
        import h5py

        self.h5_file = h5py.File(self.filepath, "r")
        weights = self.h5_file["model_weights"] if "model_weights" in self.h5_file else self.h5_file  # full model or weights only
        # return {kk: [vv[ww] for ww in vv.attrs["weight_names"]] for kk, vv in weights.items() if len(vv) > 0 and "weight_names" in vv.attrs}
        weights = {kk: weights[kk] for kk in weights.attrs['layer_names']}
        return {kk: [vv[ww] for ww in vv.attrs["weight_names"]] for kk, vv in weights.items() if len(vv) > 0}

    def __read_keras_file__(self):
        import re, json, zipfile, h5py

        archive = zipfile.ZipFile(self.filepath, "r")
        with archive.open('config.json', 'r') as ff:
            model_config = json.loads(ff.read())
        model_weight_file = archive.open('model.weights.h5', 'r')
        h5_file = h5py.File(model_weight_file, mode="r")
        self.archive, self.model_weight_file, self.h5_file = archive, model_weight_file, h5_file

        if "_layer_checkpoint_dependencies" not in h5_file:
            print(">>>> [ERROR] weights key '_layer_checkpoint_dependencies' not exists in provided keras file {}".format(self.filepath))
            return {}

        dd = {}
        used_names = {}
        weights = h5_file["_layer_checkpoint_dependencies"]
        for layer in model_config['config']['layers']:
            # from keras.src.utils import generic_utils; generic_utils.to_snake_case
            map_name = layer['class_name']
            map_name = re.sub("(.)([A-Z][a-z]+)", r"\1_\2", map_name)
            map_name = re.sub("([a-z])([A-Z])", r"\1_\2", map_name).lower()
            map_name = ("private" + map_name) if map_name[0] == "_" else map_name

            # _load_container_state
            if map_name in used_names:
                used_names[map_name] += 1
                map_name = f"{map_name}_{used_names[map_name]}"
            else:
                used_names[map_name] = 0

            # print(f"{map_name = }, {layer['name'] = }")
            if map_name in weights and "vars" in weights[map_name]:
                wws = weights[map_name]["vars"]  # This is rather slow on the first time reading [???]
                if len(wws) > 0:
                    dd[layer['name']] = list(wws.values())
                    # print("  shape:", [ii.shape for ii in dd[layer['name']]])
        return dd

    def close(self):
        self.h5_file.close()
        if self.is_keras_format:
            self.model_weight_file.close()
            self.archive.close()


def _load_layer_weights_nested_(model, weights, skip_mismatch=False, debug=False, prefix=""):
    for layer in model.layers:
        if len(layer.weights) == 0:
            continue
        if debug:
            print(">>>> Load layer weights:", "/".join([prefix, layer.name]) if prefix else layer.name, [ii.name for ii in layer.weights])
        if layer.name not in weights:
            if debug:
                print("Warning: {} not exists in provided h5 weights".format(layer.name))
            continue

        ss = weights[layer.name]
        if hasattr(layer, "layers"):  # nested
            dd = {}
            for weight_name in ss.attrs["weight_names"]:
                dd.setdefault("/".join(weight_name.split("/")[:-1]), []).append(ss[weight_name])
            _load_layer_weights_nested_(layer, dd, skip_mismatch=skip_mismatch, debug=debug, prefix=layer.name)
            continue

        if hasattr(ss, "attrs"):
            ss = [np.array(ss[ww]) for ww in ss.attrs["weight_names"]]
        elif hasattr(ss, "values"):
            ss = [np.array(ww) for ww in ss.values()]
        else:
            ss = [np.array(ww) for ww in ss]

        source_shape, target_shape = [ii.shape for ii in ss], [list(ii.shape) for ii in layer.weights]
        if debug:
            print("     [Before transpose] required weights: {}, provided: {}".format(target_shape, source_shape))
        if skip_mismatch and (len(source_shape) != len(target_shape) or any([np.prod(ii) != np.prod(jj) for ii, jj in zip(source_shape, target_shape)])):
            print("Warning: skip loading weights for layer: {}, required weights: {}, provided: {}".format(layer.name, target_shape, source_shape))
            continue

        if hasattr(layer, "set_weights_channels_last"):
            layer.set_weights_channels_last(ss)
        else:
            layer.set_weights(ss)


def load_weights_from_hdf5_file(filepath, model, skip_mismatch=False, debug=False):
    if not isinstance(filepath, str):  # Regard as dict
        _load_layer_weights_nested_(model, filepath, skip_mismatch=skip_mismatch, debug=debug)
    else:
        with H5orKerasFileReader(filepath) as weights:
            _load_layer_weights_nested_(model, weights, skip_mismatch=skip_mismatch, debug=debug)


def _save_attributes_to_hdf5_group_(group, name, data):
    """Saves attributes (data) of the specified name into the HDF5 group.

    This method deals with an inherent problem of HDF5 file which is not
    able to store data larger than HDF5_OBJECT_HEADER_LIMIT bytes.
    """
    # Check that no item in `data` is larger than `HDF5_OBJECT_HEADER_LIMIT`
    # because in that case even chunking the array would not make the saving possible.
    bad_attributes = [x for x in data if len(x) > HDF5_OBJECT_HEADER_LIMIT]

    # Expecting this to never be true.
    if bad_attributes:
        raise RuntimeError(
            "The following attributes cannot be saved to HDF5 file because " f"they are larger than {HDF5_OBJECT_HEADER_LIMIT} " f"bytes: {bad_attributes}"
        )

    data_npy = np.asarray(data)

    num_chunks = 1
    chunked_data = np.array_split(data_npy, num_chunks)

    # This will never loop forever thanks to the test above.
    while any([x.nbytes > HDF5_OBJECT_HEADER_LIMIT for x in chunked_data]):
        num_chunks += 1
        chunked_data = np.array_split(data_npy, num_chunks)

    if num_chunks > 1:
        for chunk_id, chunk_data in enumerate(chunked_data):
            group.attrs["%s%d" % (name, chunk_id)] = chunk_data
    else:
        group.attrs[name] = data


def _save_to_layer_group_nested_(layer_weights, layer_group, compression=None):
    if hasattr(layer_weights, "layers"):  # nested
        sub_model = layer_weights
        weight_names, weight_values = [], []
        for layer in sub_model.layers:
            weight_names += [ww.name.encode("utf8") for ww in layer.weights]
            weight_values += layer.get_weights_channels_last() if hasattr(layer, "get_weights_channels_last") else layer.get_weights()
        # for sub_layer in layer_weights.layers:
        #     sub_layer_group = layer_group.create_group(sub_layer.name)
        #     _save_to_layer_group_nested_(sub_layer, sub_layer_group)
        # return
    elif isinstance(layer_weights, (layers.Layer, layers.Shape)):
        layer = layer_weights
        weight_names = [ww.name.encode("utf8") for ww in layer.weights]
        weight_values = layer.get_weights_channels_last() if hasattr(layer, "get_weights_channels_last") else layer.get_weights()
    else:
        weight_names = [ww.encode("utf8") for ww in layer_weights]
        weight_values = list(layer_weights.values())

    _save_attributes_to_hdf5_group_(layer_group, "weight_names", weight_names)
    for name, val in zip(weight_names, weight_values):
        param_dset = layer_group.create_dataset(name, val.shape, dtype=val.dtype, compression=compression, chunks=True)
        if not val.shape:
            # scalar
            param_dset[()] = val
        else:
            param_dset[:] = val


def save_weights_to_hdf5_file(filepath, model, compression=None, layer_start=None, layer_end=None):
    """Saves the weights of a list of layers to a HDF5 file.
    - compression: refer `import h5py; help(h5py.File.create_dataset)`
          and `from h5py._hl import dataset; help(dataset.make_new_dset)`.
    """
    import h5py

    if isinstance(model, dict):
        weights_dict = model
    else:
        weights_dict = {layer.name: layer for layer in model.layers[layer_start:layer_end]}

    with h5py.File(filepath, "w") as h5_file:
        _save_attributes_to_hdf5_group_(h5_file, "layer_names", [layer_name.encode("utf8") for layer_name in weights_dict])
        h5_file.attrs["backend"] = backend.backend().encode("utf8")
        for layer_name, layer_weights in weights_dict.items():
            layer_group = h5_file.create_group(layer_name)
            # print(layer_name, layer_weights)
            _save_to_layer_group_nested_(layer_weights, layer_group, compression=compression)


def convert_torch_weights_to_h5(
    source_pt_path,
    save_path="AUTO",
    skip_weights=["num_batches_tracked"],
    name_convert_funcs=None,
    name_convert_map=None,
    weight_convert_funcs=None,
    to_fp16=False,
):
    """
    Examples:
    # Save weights is torch pt -> convert to h5 -> load back
    >>> os.environ['KECAM_BACKEND'] = 'torch'
    >>> from keras_cv_attention_models import download_and_load, models
    >>> mm = models.AotNet50()
    >>> mm.save('test.pt')
    >>> conv_func = lambda target_name, ww: ww.transpose([2, 3, 1, 0]) if target_name.endswith("conv") and len(ww.shape) == 4 else ww
    >>> _ = download_and_load.convert_torch_weights_to_h5('test.pt', 'test.h5', weight_convert_funcs=[conv_func])
    >>> mm.load("test.h5")
    """
    import os
    import torch

    source_state_dict = torch.load(source_pt_path, map_location=torch.device("cpu")) if isinstance(source_pt_path, str) else source_pt_path
    source_state_dict = source_state_dict.get("state_dict", source_state_dict.get("model", source_state_dict))
    name_convert_funcs = [] if name_convert_funcs is None else name_convert_funcs
    name_convert_map = {} if name_convert_map is None else name_convert_map
    weight_convert_funcs = [] if weight_convert_funcs is None else weight_convert_funcs
    save_path = (os.path.splitext(source_pt_path)[0] + ".h5" if isinstance(source_pt_path, str) else "converted.h5") if save_path == "AUTO" else save_path

    """ keras_reload_stacked_state_dict """
    stacked_state_dict = {}
    for source_name, source_weight in source_state_dict.items():
        # target_name = source_name.replace("layers.", "blocks.")
        # target_name = target_name[len("model."):] if target_name.startswith("model.") else target_name
        if any([source_name.endswith(ii) for ii in skip_weights]):
            continue
        source_name_split = source_name.split(".")
        source_layer_name, source_weight_name = ".".join(source_name_split[:-1]), source_name_split[-1]

        target_layer_name = source_layer_name
        for additional_name_func in name_convert_funcs:
            target_layer_name = additional_name_func(target_layer_name)
        for kk, vv in name_convert_map.items():
            if kk in target_layer_name:
                target_layer_name = target_layer_name.replace(kk, vv)
        print(">>>> layer {} -> {}, {}: {}".format(source_layer_name, target_layer_name, source_weight_name, source_weight.shape))

        source_weight = source_weight.numpy() if hasattr(source_weight, "numpy") else source_weight
        source_weight = source_weight.astype("float16") if to_fp16 else source_weight
        for additional_weight_func in weight_convert_funcs:
            # source_weight = [ii.T if len(ii) == 2 else ii for ii in source_weight]
            source_weight = additional_weight_func(target_layer_name, source_weight)
        print("    weight: {}, shape: {}".format(source_weight_name, source_weight.shape))
        stacked_state_dict.setdefault(target_layer_name, {}).update({source_layer_name + "/" + source_weight_name: source_weight})

    if save_path:
        print(">>>> Save to:", save_path)
        save_weights_to_hdf5_file(save_path, stacked_state_dict)
    return stacked_state_dict


""" Convert PyTorch weights """


def state_dict_stack_by_layer(state_dict, skip_weights=["num_batches_tracked"], unstack_weights=[]):
    stacked_state_dict = {}
    for kk, vv in state_dict.items():
        if kk[-2] == ":":
            kk = kk[:-2]  # Keras weight name like "..../weight:0"
        split_token = "/" if "/" in kk else "."
        split_kk = kk.split(split_token)
        vv = (vv.float().numpy() if hasattr(vv, "float") else vv.numpy()) if hasattr(vv, "numpy") else vv
        if split_kk[-1] in skip_weights:
            continue

        if split_kk[-1] in unstack_weights:
            stacked_state_dict[kk] = [vv]
        else:
            # split_kk[-1] in ["weight", "bias", "running_mean", "running_var", "gain"]
            layer_name = split_token.join(split_kk[:-1])
            stacked_state_dict.setdefault(layer_name, []).append(vv)
    return stacked_state_dict


def match_layer_names_with_torch(target_names, tail_align_dict={}, full_name_align_dict={}, tail_split_position=2):
    layer_names_matched_torch = [""] * len(target_names)
    raw_id_dict = {ii: id for id, ii in enumerate(target_names)}

    # is_tail_align_dict_split_by_stack = len(tail_align_dict) > 0 and isinstance(list(tail_align_dict.values())[0], dict)
    for id, ii in enumerate(target_names):
        name_split = ii.split("_")
        stack_name = name_split[0]
        head_name = "_".join(name_split[:tail_split_position])
        tail_name = "_".join(name_split[tail_split_position:])
        # cur_tail_align_dict = tail_align_dict[stack_name] if is_tail_align_dict_split_by_stack else tail_align_dict
        cur_tail_align_dict = tail_align_dict.get(stack_name, tail_align_dict)
        # print("id = {}, ii = {}, stack_name = {}, tail_name = {}".format(id, ii, stack_name, tail_name))
        if ii in full_name_align_dict:
            align = full_name_align_dict[ii]
            if isinstance(align, str):
                align = raw_id_dict[align]
            else:
                align = id + align if align < 0 else align
            layer_names_matched_torch.insert(align, ii)
            layer_names_matched_torch.pop(-1)
        elif tail_name in cur_tail_align_dict:
            align = cur_tail_align_dict[tail_name]
            if isinstance(align, str):
                align = raw_id_dict[head_name + "_" + align]
            else:
                align = id + align if align < 0 else align
            layer_names_matched_torch.insert(align, ii)
            layer_names_matched_torch.pop(-1)
        else:
            layer_names_matched_torch[id] = ii
    return layer_names_matched_torch


def align_layer_names_multi_stage(target_names, tail_align_dict={}, full_name_align_dict={}, tail_split_position=2, specific_match_func=None, verbose=1):
    if isinstance(tail_split_position, int):
        tail_align_dict, full_name_align_dict, tail_split_position = [tail_align_dict], [full_name_align_dict], [tail_split_position]
    full_name_align_dict = full_name_align_dict if isinstance(full_name_align_dict, (list, tuple)) else [full_name_align_dict]
    # for ii, jj, kk in zip(tail_align_dict, full_name_align_dict, tail_split_position):
    for idx in range(max(len(tail_split_position), len(full_name_align_dict))):
        cur_tail = tail_align_dict[idx] if idx < len(tail_align_dict) else {}
        cur_full = full_name_align_dict[idx] if idx < len(full_name_align_dict) else {}
        cur_split = tail_split_position[idx] if idx < len(tail_split_position) else tail_split_position[-1]
        if verbose > 0:
            print(">>>> tail_align_dict:", cur_tail)
            print(">>>> full_name_align_dict:", cur_full)
            print(">>>> tail_split_position:", cur_split)
        target_names = match_layer_names_with_torch(target_names, cur_tail, cur_full, cur_split)

    if specific_match_func is not None:
        target_names = specific_match_func(target_names)
    return target_names


def keras_reload_stacked_state_dict(model, stacked_state_dict, layer_names_matched_torch, additional_transfer={}, save_name=None, verbose=1):
    import numpy as np

    for kk, tf_layer_name in zip(stacked_state_dict.keys(), layer_names_matched_torch):
        if verbose > 0:
            print("  torch layer name: {}, tf layer name: {}".format(kk, tf_layer_name))
        tf_layer = model.get_layer(tf_layer_name)
        tf_weights = tf_layer.get_weights()
        torch_weight = stacked_state_dict[kk]
        if verbose > 0:
            print("    Before: [{}] torch: {}, tf: {}".format(kk, [ii.shape for ii in torch_weight], [ii.shape for ii in tf_weights]))

        if isinstance(tf_layer, layers.DepthwiseConv2D):  # DepthwiseConv2D is instance of Conv2D, put it first
            torch_weight[0] = np.transpose(torch_weight[0], (2, 3, 0, 1))
        elif isinstance(tf_layer, layers.Conv2D):
            torch_weight[0] = np.transpose(torch_weight[0], (2, 3, 1, 0))
            if len(torch_weight) > 2:  # gain
                torch_weight[2] = np.squeeze(torch_weight[2])
        elif isinstance(tf_layer, layers.PReLU):
            torch_weight[0] = np.expand_dims(np.expand_dims(torch_weight[0], 0), 0)
        elif isinstance(tf_layer, layers.Conv1D):
            torch_weight[0] = np.transpose(torch_weight[0], (2, 1, 0))
        elif isinstance(tf_layer, layers.Dense):
            # [Note] if it's fc layer after flatten, weights need to reshape according to NCHW --> NHWC
            torch_weight[0] = torch_weight[0].T

        for add_layer, add_transfer in additional_transfer.items():
            if isinstance(add_layer, str):
                if tf_layer.name.endswith(add_layer):
                    torch_weight = add_transfer(torch_weight)
            elif isinstance(tf_layer, add_layer):
                torch_weight = add_transfer(torch_weight)
        if verbose > 0:
            print("    After: [{}] torch: {}, tf: {}".format(kk, [ii.shape for ii in torch_weight], [ii.shape for ii in tf_weights]))

        tf_layer.set_weights(torch_weight)

    if save_name is None:
        save_name = model.name + ".h5"
    if len(save_name) != 0:
        if verbose > 0:
            print()
            print(">>>> Save model to:", save_name)
        model.save(save_name)


def try_save_pth_and_onnx(torch_model, save_pth=True, save_onnx=True, input_shape=(10, 3, 224, 224), dtype="float32", save_name=None):
    import torch
    import numpy as np

    save_name = torch_model.__class__.__name__ if save_name is None else save_name
    dummy_inputs = torch.from_numpy(np.random.uniform(size=input_shape).astype(dtype))
    if save_pth:
        output_name = save_name + ".pth"
        traced_cell = torch.jit.trace(torch_model, (dummy_inputs), strict=False)
        torch.jit.save(traced_cell, output_name)
        print(">>>> Saved to:", output_name)

    if save_onnx:
        output_name = save_name + ".onnx"
        torch.onnx.export(
            model=torch_model,
            args=dummy_inputs,
            f=output_name,
            verbose=False,
            keep_initializers_as_inputs=True,
            training=torch.onnx.TrainingMode.PRESERVE,
            do_constant_folding=False,
            opset_version=13,
        )
        print(">>>> Saved to:", output_name)


def keras_reload_from_torch_model(
    torch_model,  # Torch model, Torch state_dict wights or Torch model weights file
    keras_model=None,
    input_shape=(224, 224),
    skip_weights=["num_batches_tracked"],
    unstack_weights=[],
    tail_align_dict={},
    full_name_align_dict={},
    tail_split_position=2,
    additional_transfer={},
    specific_match_func=None,
    save_name=None,
    do_convert=True,
    do_predict=True,
    verbose=1,
):
    import torch
    import numpy as np

    from keras_cv_attention_models import test_images
    from keras_cv_attention_models.common_layers import PreprocessInput
    from keras_cv_attention_models.imagenet.eval_func import decode_predictions

    input_shape = input_shape[:2] if keras_model is None or None in keras_model.input_shape[1:] else keras_model.input_shape[1:-1]
    if isinstance(torch_model, str):
        print(">>>> Reload Torch weight file:", torch_model)
        torch_model = torch.load(torch_model, map_location=torch.device("cpu"))
        torch_model = torch_model.get("model", torch_model.get("state_dict", torch_model))
    is_state_dict = isinstance(torch_model, dict)

    """ Chelsea the cat  """
    do_predict = do_predict and do_convert
    if do_predict:
        pp = PreprocessInput(input_shape=input_shape, rescale_mode="torch")
        img = pp(test_images.cat()).numpy()

    if not is_state_dict:
        _ = torch_model.eval()
        state_dict = torch_model.state_dict()

        if do_predict:
            try:
                # from torchsummary import summary
                # summary(torch_model, (3, *input_shape), device="cpu")
                """Torch Run predict"""
                out = torch_model(torch.from_numpy(img.copy().transpose(0, 3, 1, 2).astype("float32")))
                out = out.detach().cpu().numpy()
                out = out[None] if len(out.shape) == 1 else out
                # out = tf.nn.softmax(out).numpy()  # If classifier activation is not softmax
                torch_out = decode_predictions(out)
            except Exception as error:
                print("[Error] something went wrong in running PyTorch model prediction:", error)
                pass
    else:
        state_dict = torch_model
    do_convert = do_convert and len(state_dict) > 0

    """ Convert torch weights """
    # torch_params = {kk: (np.cumproduct(vv.shape)[-1] if len(vv.shape) != 0 else 1) for kk, vv in state_dict.items() if ".num_batches_tracked" not in kk}
    stacked_state_dict = state_dict_stack_by_layer(state_dict, skip_weights=skip_weights, unstack_weights=unstack_weights)
    if verbose > 0:
        print(">>>> torch_model total_parameters :", np.sum([np.sum([np.prod(jj.shape) for jj in ii]) for ii in stacked_state_dict.values()]))
    aa = {kk: [1 if isinstance(jj, float) else jj.shape for jj in vv] for kk, vv in stacked_state_dict.items()}
    if verbose > 0:
        print(">>>> Torch weights:")
        _ = [print("  '{}': {}".format(kk, vv)) for kk, vv in aa.items()]
        print()

    if keras_model is None:
        return

    """ Keras model weights """
    target_names = [ii.name for ii in keras_model.layers if len(ii.weights) != 0]
    aa = {keras_model.get_layer(ii).name: [jj.shape.as_list() for jj in keras_model.get_layer(ii).weights] for ii in target_names}
    if verbose > 0:
        print(">>>> keras_model total_parameters :", np.sum([np.sum([int(np.prod(jj)) for jj in ii]) for ii in aa.values()]))
        print(">>>> Keras weights:")
        _ = [print("  '{}': {}".format(kk, vv)) for kk, vv in aa.items()]
        print()

    """ Load torch weights and save h5 """
    if len(tail_align_dict) != 0 or len(full_name_align_dict) != 0 or specific_match_func is not None:
        aligned_names = align_layer_names_multi_stage(target_names, tail_align_dict, full_name_align_dict, tail_split_position, specific_match_func, verbose)
        aa = {keras_model.get_layer(ii).name: [jj.shape.as_list() for jj in keras_model.get_layer(ii).weights] for ii in aligned_names}
        if verbose > 0:
            print(">>>> Keras weights matched torch:")
            _ = [print("  '{}': {}".format(kk, vv)) for kk, vv in aa.items()]
            print()
    else:
        aligned_names = target_names

    if not do_convert:
        return

    save_name = save_name if save_name is not None else keras_model.name + "_imagenet.h5"
    if verbose > 0:
        print(">>>> Keras reload torch weights:")
    keras_reload_stacked_state_dict(keras_model, stacked_state_dict, aligned_names, additional_transfer, save_name=save_name, verbose=verbose)
    if verbose > 0:
        print()

    """ Keras run predict """
    if do_predict:
        try:
            pred = keras_model(img).numpy()
            # pred = tf.nn.softmax(pred).numpy()  # If classifier activation is not softmax
            print(">>>> Keras model prediction:", decode_predictions(pred)[0])
            print()
            if not is_state_dict:
                print(">>>> Torch model prediction:", torch_out)
        except:
            pass
